# -*- coding: utf-8 -*-
"""
Created on Tue Mar 17 20:21:05 2020

@author: Peter M. Clausen
"""

import glob
import re

import ford.reader as reader
from ford.reader import _contains_unterminated_string
import pytest

from conftest import gfortran_is_not_installed

RE_WHITE = re.compile(r"\s+")


def remove_multiple_white_space(lines):
    return [RE_WHITE.sub(" ", line.strip()) for line in lines]


def test_reader_test_data():
    """Basic regression test"""
    f_files = glob.glob("./test_data/*.f*")
    f_files = [
        f
        for f in f_files
        if "expected" not in f and "bad" not in f  # remove 'expected' and 'bad' files
    ]
    # Set to True to update the 'expected' files
    create_expected = False
    for ff in f_files:
        ee = ff.replace(".f90", "_expected.f90")
        print("\tProcessing: %s \tExpected: %s " % (ff, ee))
        lines = [
            line
            for line in reader.FortranReader(
                ff, docmark="!", predocmark=">", docmark_alt="#", predocmark_alt="<"
            )
        ]
        if create_expected:
            print("WARNING : Writing expected file " + ee)
            with open(ee, "w") as ef:
                for line in lines:
                    print(line)
                    ef.write(line + "\n")
        else:
            with open(ee) as ef:
                lines = remove_multiple_white_space(lines)
                elines = remove_multiple_white_space(ef.readlines())
                assert lines == elines


def test_reader_continuation(copy_fortran_file):
    """Checks that line continuations are handled correctly"""

    data = """\
    program foo
    !! some docs
    integer :: bar = &
    &
    4
    end
    """

    filename = copy_fortran_file(data)

    lines = list(reader.FortranReader(filename, docmark="!"))
    assert lines == ["program foo", "!! some docs", "integer :: bar = 4", "end"]


def test_type(copy_fortran_file):
    """Check that types can be read"""

    data = """\
    program foo
    !! base type
    type :: base
    end type base

    !! derived type
    type, extends(base) :: derived
    end type
    """

    expected = [
        "program foo",
        "!! base type",
        "type :: base",
        "end type base",
        "!! derived type",
        "type, extends(base) :: derived",
        "end type",
    ]

    filename = copy_fortran_file(data)

    lines = list(reader.FortranReader(filename, docmark="!"))
    assert lines == expected


def test_unknown_include(copy_fortran_file):
    """Check that `include "file.h"` ignores unknown files"""

    data = """\
    program test
    include "file.h"
    end program test
    """

    expected = [
        "program test",
        'include "file.h"',
        "end program test",
    ]

    filename = copy_fortran_file(data)

    lines = list(reader.FortranReader(filename, docmark="!"))
    assert lines == expected


def test_unterminated_strings():
    """Check the utility function works"""
    assert _contains_unterminated_string(""" bad "quote """)
    assert _contains_unterminated_string(""" bad 'quote """)
    assert _contains_unterminated_string(""" bad "'quote """)
    assert _contains_unterminated_string(""" bad 'quote" """)
    assert _contains_unterminated_string(""" "multiple" bad "'quote """)
    assert _contains_unterminated_string(""" bad 'quote" """)

    assert not _contains_unterminated_string(""" good "quote" """)
    assert not _contains_unterminated_string(""" good 'quote' """)
    assert not _contains_unterminated_string(""" good "'quote" """)
    assert not _contains_unterminated_string(""" good 'quote"' """)
    assert not _contains_unterminated_string(""" "multiple" good "'quote" """)
    assert not _contains_unterminated_string(""" good 'quote"' """)


def test_multiline_string(copy_fortran_file):
    """Check that we can continue string literals including exclamation
    marks over multiple lines. Issue #320"""

    data = '''\
    program multiline_string
      implicit none
      print*, 'dont''t', " get ""!>quotes!@""", " '""!<wrong!!""' ", "foo&
              &bar! foo&
              &bar"
    end program multiline_string
    '''

    filename = copy_fortran_file(data)
    expected = [
        "program multiline_string",
        "implicit none",
        '''print*, 'dont''t', " get ""!>quotes!@""", " '""!<wrong!!""' ", "foobar! foobar"''',
        "end program multiline_string",
    ]

    lines = list(
        reader.FortranReader(
            filename, docmark="!", predocmark="<", docmark_alt=">", predocmark_alt="@"
        )
    )
    assert lines == expected


@pytest.mark.skipif(
    gfortran_is_not_installed(), reason="Requires gfortran to be installed"
)
def test_preprocessor(copy_fortran_file):
    """Check some basic preprocessing is applied"""

    data = """\
    subroutine SUB_NAME()
    end sub_name SUB_NAME
    """

    filename = copy_fortran_file(data, ".F90")

    lines = "\n".join(
        list(
            reader.FortranReader(
                str(filename),
                preprocessor=["cpp", "-traditional-cpp", "-E"],
                macros=["SUB_NAME=foo"],
            )
        )
    )
    assert "foo" in lines


@pytest.mark.skipif(
    gfortran_is_not_installed(), reason="Requires gfortran to be installed"
)
def test_preprocessor_warning(copy_fortran_file):
    """Check preprocessing is still done even if there are warnings"""

    data = """\
    #warning something
    subroutine SUB_NAME()
    end sub_name SUB_NAME
    """

    filename = copy_fortran_file(data, ".F90")

    lines = "\n".join(
        list(
            reader.FortranReader(
                str(filename),
                preprocessor=["cpp", "-traditional-cpp", "-E"],
                macros=["SUB_NAME=foo"],
            )
        )
    )
    assert "foo" in lines



import pathlib
import re
import sys
from typing import List

import ford
import ford.fortran_project

from bs4 import BeautifulSoup

# Ford default src folder
DEFAULT_SRC = "src"


def run_ford(monkeypatch, md_file: pathlib.Path, extra_args: list = None):
    """Modify command line args with argv"""
    with monkeypatch.context() as m:
        command = ["ford", str(md_file)]
        if extra_args is not None:
            command = command[:1] + extra_args + command[1:]
        m.setattr(sys, "argv", command)
        ford.run()


def test_extra_mods_empty(
    copy_fortran_file,
    copy_settings_file,
    monkeypatch,
    restore_macros,
    restore_nameselector,
):
    """This checks that extra_mods is parsed correctly in input md file"""

    data = """\
    module test
    end module test
    """
    settings = """\
    search: false
    extra_mods:
    """

    copy_fortran_file(data)
    md_file = copy_settings_file(settings)

    run_ford(monkeypatch, md_file)


def test_extra_mods_intrinsic(
    copy_fortran_file,
    copy_settings_file,
    monkeypatch,
    restore_macros,
    restore_nameselector,
):
    """This checks that adding extra_mods doesn't change the module variable INTRINSIC_MODS"""
    data = """\
    module test
    end module test
    """
    settings = """\
    search: false
    extra_mods: dummy: dummy_module
    """
    # Initial value of intrinsic mods
    old_intrinsic_mods = ford.fortran_project.INTRINSIC_MODS.copy()

    # set up project data
    copy_fortran_file(data)
    md_file = copy_settings_file(settings)

    run_ford(monkeypatch, md_file)

    # Check that the module level variable has not been changed
    assert ford.fortran_project.INTRINSIC_MODS == old_intrinsic_mods


def get_main_body_text(html_dir: pathlib.Path, filename: str) -> List[str]:
    with open(html_dir / filename, "r") as f:
        index_html = BeautifulSoup(f.read(), features="html.parser")
    return index_html.find_all(string=re.compile("Test:"))


def test_default_aliases(
    copy_fortran_file,
    copy_settings_file,
    tmp_path,
    monkeypatch,
    restore_macros,
    restore_nameselector,
):
    """Check that aliases specified in project file are replaced correctly"""

    data = """\
    module test
    !! Test: The project media url (|media|) should work here too
    end module test
    """
    settings = """\
    search: false

    Test: The project media url should be |media|
    """

    copy_fortran_file(data)
    md_file = copy_settings_file(settings)

    run_ford(monkeypatch, md_file)

    html_dir = tmp_path / "doc"
    index_text = get_main_body_text(html_dir, "index.html")
    expected_index_text = [
        "Test: The project media url should be ./media",
    ]
    assert index_text == expected_index_text

    module_text = get_main_body_text(html_dir, "module/test.html")
    expected_module_text = [
        "Test: The project media url (./media) should work here too"
    ]
    assert module_text == expected_module_text


def test_one_alias(
    copy_fortran_file,
    copy_settings_file,
    tmp_path,
    monkeypatch,
    restore_macros,
    restore_nameselector,
):
    """Check that aliases specified in project file are replaced correctly"""

    data = """\
    module test
    !! Title
    !!
    !! Test: This foo should not be exanded
    !!
    !! Test: But this |foo| should be 'bar'
    end module test
    """
    settings = """\
    search: false
    alias: foo = bar

    Test: This |foo| should be expanded as 'bar'

    Test: This foo should not be.
    """

    copy_fortran_file(data)
    md_file = copy_settings_file(settings)

    run_ford(monkeypatch, md_file)

    html_dir = tmp_path / "doc"
    paragraphs = get_main_body_text(html_dir, "index.html")
    expected_paragraphs = [
        "Test: This bar should be expanded as 'bar'",
        "Test: This foo should not be.",
    ]
    assert paragraphs == expected_paragraphs

    module_text = get_main_body_text(html_dir, "module/test.html")
    expected_module_text = [
        "Test: This foo should not be exanded",
        "Test: But this bar should be 'bar'",
    ]
    assert module_text == expected_module_text


def test_multiple_aliases(
    copy_fortran_file,
    copy_settings_file,
    tmp_path,
    monkeypatch,
    restore_macros,
    restore_nameselector,
):
    """Check that aliases specified in project file are replaced correctly"""

    data = """\
    module test
    !! Title
    !!
    !! Test: This foo and zing should not be exanded
    !!
    !! Test: But this |foo| should be 'bar', and this |zing| 'quaff'
    end module test
    """
    settings = """\
    search: false
    alias: foo = bar
           zing = quaff

    Test: This |foo| should be expanded as 'bar', while |zing| should be 'quaff'

    Test: This foo and zing should not be.
    """

    copy_fortran_file(data)
    md_file = copy_settings_file(settings)

    run_ford(monkeypatch, md_file)

    html_dir = tmp_path / "doc"
    paragraphs = get_main_body_text(html_dir, "index.html")
    expected_paragraphs = [
        "Test: This bar should be expanded as 'bar', while quaff should be 'quaff'",
        "Test: This foo and zing should not be.",
    ]

    assert paragraphs == expected_paragraphs

    module_text = get_main_body_text(html_dir, "module/test.html")
    expected_module_text = [
        "Test: This foo and zing should not be exanded",
        "Test: But this bar should be 'bar', and this quaff 'quaff'",
    ]
    assert module_text == expected_module_text



# -*- coding: utf-8 -*-
#
#  md_environ.py
#
#  Copyright 2015 Christopher MacMackin <cmacmackin@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from os import getenv

from markdown import Extension
from markdown.inlinepatterns import Pattern

ENVIRON_RE = r"\$\{(\w+)\}"


class EnvironPattern(Pattern):
    """
    Pattern to pick out environment variables and insert their value.
    """

    def handleMatch(self, m):
        var = m.group(2)
        return getenv(var, "")


def makeExtension(**kwargs):
    """Inform Markdown of the existence of the extension."""
    return EnvironExtension(**kwargs)


class EnvironExtension(Extension):
    """
    Extension: ${VARNAME} will be replaced by contents of environment
    variable VARNAME, it it is defined, or by an empty string otherwise.
    """

    def extendMarkdown(self, md, *args, **kwargs):
        """Insert 'environ' pattern before 'not_strong' pattern."""
        md.inlinePatterns.register(EnvironPattern(ENVIRON_RE), "environ", 65)
        md.registerExtension(self)



import os
import markdown

from ford.md_environ import EnvironExtension


def test_md_environ():
    md = markdown.Markdown(
        extensions=[EnvironExtension()],
        output_format="html5",
    )

    expected_text = "success"
    os.environ["FORD_TEST_ENV_VAR"] = expected_text

    converted = md.convert("${FORD_TEST_ENV_VAR}")

    assert converted == f"<p>{expected_text}</p>"



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from importlib.metadata import version as get_version

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "FORD"
copyright = "2022, Chris MacMackin, FORD Contributors"
author = "Chris MacMackin, Peter Hill"

# The full version, including alpha/beta/rc tags
release = get_version(project)
# Major.minor version
version = ".".join(release.split(".")[:2])

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx_argparse_cli",
]

# Napoleon settings
napoleon_include_init_with_doc = True
napoleon_attr_annotations = True

templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The default role for text marked up `like this`
default_role = "any"

# Tell sphinx what the primary language being documented is.
primary_domain = "py"

# Tell sphinx what the pygments highlight language should be.
highlight_language = "fortran"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
html_static_path = ["_static"]

html_context = {
    "github_user": "Fortran-FOSS-Programmers",
    "github_repo": "ford",
    "github_version": "master",
    "doc_path": "doc",
}

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = dict(
    # analytics_id=''  this is configured in rtfd.io
    # canonical_url="",
    repository_url="https://github.com/Fortran-FOSS-Programmers/ford",
    repository_branch="master",
    path_to_docs="docs",
    use_edit_page_button=True,
    use_repository_button=True,
    use_issues_button=True,
    home_page_in_toc=False,
)

# -- Options for intersphinx extension ---------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#configuration

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}



from ford.md_striped_table import StripedTableCSSExtension
import markdown
from bs4 import BeautifulSoup
import textwrap


def test_zebra_table():
    """Check that alternate rows have different classes"""
    md = markdown.Markdown(
        extensions=["markdown.extensions.extra", StripedTableCSSExtension()],
        output_format="html5",
    )

    text = textwrap.dedent(
        """
        | Header1    | Header2     |
        | ---------- | ----------- |
        | some text  | some text   |
        | some text  | some text   |
        | some text  | some text   |
        | some text  | some text   |
        """
    )

    html = md.convert(text)
    soup = BeautifulSoup(html, features="html.parser")

    print(soup)
    assert "table" in soup.table.attrs["class"]
    assert "table-striped" in soup.table.attrs["class"]


def test_no_clobber_html_table():
    """Check that tables with CSS already applied don't get modified"""
    md = markdown.Markdown(
        extensions=["markdown.extensions.extra", StripedTableCSSExtension()],
        output_format="html5",
    )

    text = textwrap.dedent(
        """
        <table class="dont-touch">
        <thead>
        <tr>
        <th>Header1</th>
        <th>Header2</th>
        </tr>
        </thead>
        <tbody>
        <tr class="row1">
        <td>some text</td>
        <td>some text</td>
        </tr>
        <tr class="row2">
        <td>some text</td>
        <td>some text</td>
        </tr>
        <tr class="row3">
        <td>some text</td>
        <td>some text</td>
        </tr>
        <tr class="row4">
        <td>some text</td>
        <td>some text</td>
        </tr>
        </tbody>
        </table>
        """
    )

    html = md.convert(text)
    soup = BeautifulSoup(html, features="html.parser")

    print(soup)
    assert soup.table.attrs["class"] == ["dont-touch"]

    body = soup.table.find("tbody")
    rows = body.find_all("tr")
    assert rows[0].attrs["class"] == ["row1"]
    assert rows[1].attrs["class"] == ["row2"]
    assert rows[2].attrs["class"] == ["row3"]
    assert rows[3].attrs["class"] == ["row4"]



from ford.pagetree import PageNode, get_page_tree

from markdown import Markdown

from textwrap import dedent


def test_footnotes_on_one_page(tmp_path):
    """Check that footnotes only appear on the pages they are defined on,
    issue #327"""

    with open(tmp_path / "index.md", "w") as f:
        f.write("title: Index")

    with open(tmp_path / "a.md", "w") as f:
        f.write(
            dedent(
                """\
            title: Page A

            This has a footnote[^1] that should only appear on this page

            [^1]: This is the footnote on page A
            """
            )
        )

    with open(tmp_path / "b.md", "w") as f:
        f.write(
            dedent(
                """\
            title: Page B

            This page should not have any footnotes
            """
            )
        )

    md = Markdown(extensions=["markdown.extensions.meta", "markdown.extensions.extra"])
    result_dir = tmp_path / "result"
    nodes = get_page_tree(tmp_path, result_dir, md)

    assert len(nodes.subpages) == 2
    assert "This is the footnote on page A" in nodes.subpages[0].contents
    assert "This is the footnote on page A" not in nodes.subpages[1].contents


def test_footnotes_on_one_page_parse_failure(tmp_path):
    """Check that footnotes only appear on the pages they are defined on,
    even if we fail to parse one of the files, issue #327"""

    with open(tmp_path / "index.md", "w") as f:
        f.write("title: Index")

    with open(tmp_path / "a.md", "w") as f:
        f.write(
            dedent(
                """\
            This has a footnote[^1] that should only appear on this page

            [^1]: This is the footnote on page A
            """
            )
        )

    with open(tmp_path / "b.md", "w") as f:
        f.write(
            dedent(
                """\
            title: Page B

            This page should not have any footnotes
            """
            )
        )

    md = Markdown(extensions=["markdown.extensions.meta", "markdown.extensions.extra"])
    result_dir = tmp_path / "result"
    nodes = get_page_tree(tmp_path, result_dir, md)

    assert len(nodes.subpages) == 1
    assert "This is the footnote on page A" not in nodes.subpages[0].contents



import ford

import subprocess
import os
import sys

import pytest


@pytest.mark.slow
@pytest.mark.parametrize(
    ("url", "config_file", "flags"),
    [
        (
            "https://bitbucket.org/gyrokinetics/gs2",
            "gs2/docs/docs.config.md",
            '--macro FCOMPILER=_GFORTRAN_ --macro GIT_HASH="a0efef2c9a499e2516e2bca7cf17e372741b96ea" --macro GIT_HASH_MAKEFILES="84aca9698fb00defc74d38ad6395a557114a9e76" --macro GIT_HASH_UTILS="0c8148c28064f5fc5025a3b1c6fff39d520b1c0b" --macro GIT_VERSION="8.1.2-327-ga0efef2c" --macro GIT_BRANCH="master" --macro MPI --macro ISO_C_BINDING --macro NEW_DIAG --macro FFT=_FFTW3_ --macro NETCDF --macro GK_NETCDF_DEFAULT_COMPRESSION_ON --macro F200X_INTRINSICS --macro SPFUNC=_SPF200X_ --macro GK_HAS_COMPILER_OPTIONS_2008 --macro WITH_EIG --macro HAVE_MPI --macro FORTRAN_NETCDF --macro ISO_C_BINDING -r a0efef2c9a499e2516e2bca7cf17e372741b96ea',  # noqa E501
        ),
        (
            "https://github.com/D3DEnergetic/FIDASIM",
            "FIDASIM/docs/fidasim.md",
            "-d ../src -d ../tables -d ../lib/idl -d ../lib/python/fidasim -p ../docs/user-guide -o ../docs/html",
        ),
        ("https://github.com/QcmPlab/HoneyTools", "HoneyTools/docs.config", ""),
        (
            "https://github.com/cibinjoseph/naturalFRUIT",
            "naturalFRUIT/ford_input.md",
            "",
        ),
        ("https://github.com/fortran-lang/fftpack", "fftpack/API-doc-FORD-file.md", ""),
        ("https://github.com/fortran-lang/fpm", "fpm/docs.md", ""),
        ("https://github.com/fortran-lang/stdlib", "stdlib/API-doc-FORD-file.md", ""),
        (
            "https://github.com/jacobwilliams/Fortran-Astrodynamics-Toolkit",
            "Fortran-Astrodynamics-Toolkit/fortran-astrodynamics-toolkit.md",
            "",
        ),
        (
            "https://github.com/jacobwilliams/bspline-fortran",
            "bspline-fortran/bspline-fortran.md",
            "",
        ),
        ("https://github.com/jacobwilliams/dvode", "dvode/ford.md", ""),
        (
            "https://github.com/jacobwilliams/fortran-csv-module",
            "fortran-csv-module/fortran-csv-module.md",
            "",
        ),
        (
            "https://github.com/jacobwilliams/json-fortran",
            "json-fortran/json-fortran.md",
            "",
        ),
        (
            "https://github.com/jacobwilliams/pyplot-fortran",
            "pyplot-fortran/pyplot-fortran.md",
            "",
        ),
        ("https://github.com/jacobwilliams/quadpack", "quadpack/quadpack.md", ""),
        ("https://github.com/szaghi/FLAP", "FLAP/doc/main_page.md", ""),
        ("https://github.com/szaghi/FiNeR", "FiNeR/doc/main_page.md", ""),
        ("https://github.com/szaghi/VTKFortran", "VTKFortran/doc/main_page.md", ""),
        ("https://github.com/toml-f/toml-f", "toml-f/docs.md", ""),
        ("https://github.com/ylikx/forpy", "forpy/forpy_project.md", ""),
        # The following all specify markdown extension config in a
        # non-supported way
        # (
        #     "https://github.com/Fortran-FOSS-Programmers/FOODIE",
        #     "FOODIE/doc/main_page.md",
        #     "--debug",
        # ),
        # (
        #     "https://github.com/Fortran-FOSS-Programmers/FoXy",
        #     "FoXy/doc/main_page.md",
        #     "--debug",
        # ),
        # ("https://github.com/szaghi/HASTY", "HASTY/doc/main_page.md", ""),
        # ("https://github.com/szaghi/BeFoR64", "BeFoR64/doc/main_page.md", ""),
        # ("https://github.com/szaghi/FURY", "FURY/doc/main_page.md", ""),
        # Source files require fypp
        # ("https://github.com/cp2k/dbcsr", "dbcsr/DBCSR.md", "--debug"),
        # output_dir subdirectory of src_dir, requires manually deleting
        # (
        #     "https://github.com/cibinjoseph/C81-Interface",
        #     "C81-Interface/ford_input.md",
        #     "-p ford_input",
        # ),
    ],
)
def test_copus(tmp_path, url, config_file, flags):
    os.chdir(tmp_path)

    subprocess.run(f"git clone --depth=1 {url}", check=True, shell=True)

    with pytest.MonkeyPatch.context() as m:
        command = f"ford {flags} {config_file}"
        m.setattr(sys, "argv", command.split())
        ford.run()



# -*- coding: utf-8 -*-
#
#  output.py
#  This file is part of FORD.
#
#  Copyright 2015 Christopher MacMackin <cmacmackin@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import sys
import os
import shutil
import traceback
from itertools import chain
import pathlib

import jinja2

from tqdm import tqdm

import ford.sourceform
import ford.tipue_search
import ford.utils
from ford.graphs import graphviz_installed, GraphManager

loc = pathlib.Path(__file__).parent
env = jinja2.Environment(
    loader=jinja2.FileSystemLoader(loc / "templates"),
    trim_blocks=True,
    lstrip_blocks=True,
)
env.globals["path"] = os.path  # this lets us call path.* in templates


def is_more_than_one(collection):
    return collection > 1


env.tests["more_than_one"] = is_more_than_one

USER_WRITABLE_ONLY = 0o755


class Documentation(object):
    """
    Represents and handles the creation of the documentation files from
    a project.
    """

    def __init__(self, data, proj_docs, project, pagetree):
        # This lets us use meta data anywhere within the template.
        # Also, in future for other template, we may not need to
        # pass the data obj.
        env.globals["projectData"] = data
        self.project = project
        # Jinja2's `if` statement counts `None` as truthy, so to avoid
        # lots of refactoring and messiness in the templates, just get
        # rid of None values
        self.data = {k: v for k, v in data.items() if v is not None}
        self.lists = []
        self.docs = []
        self.njobs = int(self.data["parallel"])
        self.parallel = self.njobs > 0

        self.index = IndexPage(self.data, project, proj_docs)
        self.search = SearchPage(self.data, project)
        if not graphviz_installed and data["graph"]:
            print(
                "Warning: Will not be able to generate graphs. Graphviz not installed."
            )
        if self.data["relative"]:
            graphparent = "../"
        else:
            graphparent = ""
        print("Creating HTML documentation...")
        try:
            # Create individual entity pages
            entity_list_page_map = [
                (project.types, TypePage),
                (project.absinterfaces, AbsIntPage),
                (project.procedures, ProcPage),
                (project.submodprocedures, ProcPage),
                (project.modules, ModulePage),
                (project.submodules, ModulePage),
                (project.programs, ProgPage),
                (project.blockdata, BlockPage),
            ]
            if data["incl_src"]:
                entity_list_page_map.append((project.allfiles, FilePage))

            for entity_list, page_class in entity_list_page_map:
                for item in entity_list:
                    self.docs.append(page_class(self.data, project, item))

            # Create lists of each entity type
            if len(project.procedures) > 0:
                self.lists.append(ProcList(self.data, project))
            if data["incl_src"] and (len(project.files) + len(project.extra_files) > 1):
                self.lists.append(FileList(self.data, project))
            if len(project.modules) + len(project.submodules) > 0:
                self.lists.append(ModList(self.data, project))
            if len(project.programs) > 1:
                self.lists.append(ProgList(self.data, project))
            if len(project.types) > 0:
                self.lists.append(TypeList(self.data, project))
            if len(project.absinterfaces) > 0:
                self.lists.append(AbsIntList(self.data, project))
            if len(project.blockdata) > 1:
                self.lists.append(BlockList(self.data, project))

            # Create static pages
            self.pagetree = [
                PagetreePage(self.data, project, item) for item in (pagetree or [])
            ]
        except Exception:
            if data["dbg"]:
                traceback.print_exc()
                sys.exit("Error encountered.")
            else:
                sys.exit('Error encountered. Run with "--debug" flag for traceback.')

        self.graphs = GraphManager(
            self.data["project_url"],
            self.data["output_dir"],
            self.data.get("graph_dir", ""),
            graphparent,
            self.data["coloured_edges"],
            self.data["show_proc_parent"],
            save_graphs=bool(self.data.get("graph_dir", False)),
        )

        if graphviz_installed and data["graph"]:
            for entity_list in [
                project.types,
                project.procedures,
                project.submodprocedures,
                project.modules,
                project.submodules,
                project.programs,
                project.files,
                project.blockdata,
            ]:
                for item in entity_list:
                    self.graphs.register(item)

            self.graphs.graph_all()
            project.callgraph = self.graphs.callgraph
            project.typegraph = self.graphs.typegraph
            project.usegraph = self.graphs.usegraph
            project.filegraph = self.graphs.filegraph
        else:
            project.callgraph = ""
            project.typegraph = ""
            project.usegraph = ""
            project.filegraph = ""

        if data["search"]:
            url = "" if data["relative"] else data["project_url"]
            self.tipue = ford.tipue_search.Tipue_Search_JSON_Generator(
                data["output_dir"], url
            )
            self.tipue.create_node(self.index.html, "index.html", {"category": "home"})
            jobs = len(self.docs) + len(self.pagetree)
            for p in tqdm(
                chain(self.docs, self.pagetree),
                total=jobs,
                unit="",
                desc="Creating search index",
            ):
                self.tipue.create_node(p.html, p.loc, p.meta)
            print("")

    def writeout(self):
        out_dir: pathlib.Path = self.data["output_dir"]
        print(f"Writing documentation to '{out_dir}'...")
        # Remove any existing file/directory. This avoids errors coming from
        # `shutils.copytree` for Python < 3.8, where we can't explicitly ignore them
        if out_dir.is_file():
            out_dir.unlink()
        else:
            shutil.rmtree(out_dir, ignore_errors=True)

        try:
            out_dir.mkdir(USER_WRITABLE_ONLY, parents=True)
        except Exception as e:
            print(f"Error: Could not create output directory. {e.args[0]}")

        for directory in [
            "lists",
            "sourcefile",
            "type",
            "proc",
            "interface",
            "module",
            "program",
            "src",
            "blockdata",
        ]:
            (out_dir / directory).mkdir(USER_WRITABLE_ONLY)

        for directory in ["css", "fonts", "js"]:
            copytree(loc / directory, out_dir / directory)

        if self.data["graph"]:
            self.graphs.output_graphs(self.njobs)
        if self.data["search"]:
            copytree(loc / "tipuesearch", out_dir / "tipuesearch")
            self.tipue.print_output()

        try:
            copytree(self.data["media_dir"], out_dir / "media")
        except OSError as e:
            print(
                f"Warning: error copying media directory {self.data['media_dir']}, {e}"
            )
        except KeyError:
            pass

        if "css" in self.data:
            shutil.copy(self.data["css"], out_dir / "css" / "user.css")

        if self.data["favicon"] == "default-icon":
            favicon_path = loc / "favicon.png"
        else:
            favicon_path = self.data["favicon"]

        shutil.copy(favicon_path, out_dir / "favicon.png")

        if self.data["incl_src"]:
            for src in self.project.allfiles:
                shutil.copy(src.path, out_dir / "src" / src.name)

        if "mathjax_config" in self.data:
            mathjax_path = out_dir / "js" / "MathJax-config"
            mathjax_path.mkdir(parents=True, exist_ok=True)
            shutil.copy(
                self.data["mathjax_config"],
                mathjax_path / os.path.basename(self.data["mathjax_config"]),
            )

        for p in chain(self.docs, self.lists, self.pagetree, [self.index, self.search]):
            p.writeout()

        print(f"\nBrowse the generated documentation: file://{out_dir}/index.html")


class BasePage:
    """
    Abstract class for representation of pages in the documentation.

      data
        Dictionary containing project information (to be used when rendering)
      proj
        FortranProject object
      obj
        The object/item in the code which this page is documenting
    """

    def __init__(self, data, proj, obj=None):
        self.data = data
        self.proj = proj
        self.obj = obj
        self.meta = getattr(obj, "meta", {})
        self.out_dir = self.data["output_dir"]
        self.page_dir = self.out_dir / "page"

    @property
    def html(self):
        """Wrapper for only doing the rendering on request (drastically reduces memory)"""
        try:
            return self.render(self.data, self.proj, self.obj)
        except Exception as e:
            raise RuntimeError(
                f"Error rendering '{self.outfile.name}' for '{self.obj.name}' : {e}"
            )

    def writeout(self):
        with open(self.outfile, "wb") as out:
            out.write(self.html.encode("utf8"))

    def render(self, data, proj, obj):
        """
        Get the HTML for the page. This method must be overridden. Arguments
        are proj_data, project object, and item in the code which the
        page documents.
        """
        raise NotImplementedError("Should not instantiate BasePage type")


class ListTopPage(BasePage):
    @property
    def list_page(self):
        raise NotImplementedError("ListTopPage subclass missing 'list_page' property")

    @property
    def outfile(self):
        return self.out_dir / self.list_page

    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = "."
            ford.sourceform.set_base_url(".")
            ford.pagetree.set_base_url(".")
        template = env.get_template(self.list_page)
        return template.render(data, project=proj, proj_docs=obj)


class IndexPage(ListTopPage):
    list_page = "index.html"


class SearchPage(ListTopPage):
    list_page = "search.html"


class ListPage(BasePage):
    @property
    def out_page(self):
        raise NotImplementedError("ListPage subclass missing 'out_page' property")

    @property
    def list_page(self):
        raise NotImplementedError("ListPage subclass missing 'list_page' property")

    @property
    def outfile(self):
        return self.out_dir / "lists" / self.out_page

    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        template = env.get_template(self.list_page)
        return template.render(data, project=proj)


class ProcList(ListPage):
    out_page = "procedures.html"
    list_page = "proc_list.html"


class FileList(ListPage):
    out_page = "files.html"
    list_page = "file_list.html"


class ModList(ListPage):
    out_page = "modules.html"
    list_page = "mod_list.html"


class ProgList(ListPage):
    out_page = "programs.html"
    list_page = "prog_list.html"


class TypeList(ListPage):
    out_page = "types.html"
    list_page = "types_list.html"


class AbsIntList(ListPage):
    out_page = "absint.html"
    list_page = "absint_list.html"


class BlockList(ListPage):
    out_page = "blockdata.html"
    list_page = "block_list.html"


class DocPage(BasePage):
    """
    Abstract class to be inherited by all pages for items in the code.
    """

    @property
    def page_path(self):
        raise NotImplementedError("DocPage subclass missing 'page_path'")

    @property
    def payload_key(self):
        raise NotImplementedError("DocPage subclass missing 'payload_key'")

    @property
    def object_page(self):
        return self.obj.ident + ".html"

    @property
    def loc(self):
        return pathlib.Path(self.obj.get_dir()) / self.object_page

    @property
    def outfile(self):
        return self.out_dir / self.obj.get_dir() / self.object_page

    def render(self, data, project, object):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        template = env.get_template(self.page_path)
        try:
            return template.render(data, project=project, **{self.payload_key: object})
        except jinja2.exceptions.TemplateError:
            print(f"Error rendering page '{self.outfile}'")
            raise


class FilePage(DocPage):
    page_path = "file_page.html"
    payload_key = "src"


class TypePage(DocPage):
    page_path = "type_page.html"
    payload_key = "dtype"


class AbsIntPage(DocPage):
    page_path = "nongenint_page.html"
    payload_key = "interface"


class ModulePage(DocPage):
    page_path = "mod_page.html"
    payload_key = "module"


class ProgPage(DocPage):
    page_path = "prog_page.html"
    payload_key = "program"


class BlockPage(DocPage):
    page_path = "block_page.html"
    payload_key = "blockdat"


class ProcedurePage(DocPage):
    page_path = "proc_page.html"
    payload_key = "procedure"


class GenericInterfacePage(DocPage):
    page_path = "genint_page.html"
    payload_key = "interface"


class InterfacePage(DocPage):
    page_path = "nongenint_page.html"
    payload_key = "interface"


def ProcPage(data, proj, obj):
    """Factory function for creating procedure or interface pages"""
    if obj.obj == "proc":
        return ProcedurePage(data, proj, obj)
    if obj.generic:
        return GenericInterfacePage(data, proj, obj)
    return InterfacePage(data, proj, obj)


class PagetreePage(BasePage):
    @property
    def object_page(self):
        return self.obj.filename + ".html"

    @property
    def loc(self):
        return pathlib.Path("page") / self.obj.location / self.object_page

    @property
    def outfile(self):
        return self.page_dir / self.obj.location / self.object_page

    def render(self, data, proj, obj):
        if data["relative"]:
            base_url = ("../" * len(obj.hierarchy))[:-1]
            if obj.filename == "index":
                if len(obj.hierarchy) > 0:
                    base_url = base_url + "/.."
                else:
                    base_url = ".."
            ford.sourceform.set_base_url(base_url)
            ford.pagetree.set_base_url(base_url)
            data["project_url"] = base_url
        template = env.get_template("info_page.html")
        obj.contents = ford.utils.sub_links(
            ford.utils.sub_macros(ford.utils.sub_notes(obj.contents)), proj
        )
        return template.render(data, page=obj, project=proj, topnode=obj.topnode)

    def writeout(self):
        if self.obj.filename == "index":
            (self.page_dir / self.obj.location).mkdir(USER_WRITABLE_ONLY, exist_ok=True)
        super(PagetreePage, self).writeout()

        for item in self.obj.copy_subdir:
            item_path = self.data["page_dir"] / self.obj.location / item
            try:
                copytree(item_path, self.page_dir / self.obj.location / item)
            except Exception as e:
                print(
                    f"Warning: could not copy directory '{item_path}'. Error: {e.args[0]}"
                )

        for item in self.obj.files:
            item_path = self.data["page_dir"] / self.obj.location / item
            try:
                shutil.copy(item_path, self.page_dir / self.obj.location)
            except Exception as e:
                print(f"Warning: could not copy file '{item_path}'. Error: {e.args[0]}")


def copytree(src: pathlib.Path, dst: pathlib.Path) -> None:
    """Wrapper around `shutil.copytree` that:
    a) doesn't try to set xattrs; and
    b) ensures modification time is time of current FORD run
    """
    shutil.copytree(src, dst, copy_function=shutil.copy)
    for file in dst.rglob("*"):
        file.touch()



import pytest

import ford


def test_sub_macro(restore_macros):
    ford.utils.register_macro("a=b")
    result = ford.utils.sub_macros("|a|")

    assert result == "b"


def test_sub_macro_with_equals(restore_macros):
    ford.utils.register_macro("a=b=c")
    result = ford.utils.sub_macros("|a|")

    assert result == "b=c"


def test_register_macro_clash(restore_macros):
    ford.utils.register_macro("a=b")

    # Should be ok registering the same key-value pair again
    ford.utils.register_macro("a=b")

    with pytest.raises(RuntimeError):
        ford.utils.register_macro("a=c")


@pytest.mark.parametrize("string", ["true", "True", "TRUE", "tRuE"])
def test_str_to_bool_true(string):
    assert ford.utils.str_to_bool(string)


@pytest.mark.parametrize("string", ["false", "False", "FALSE", "fAlSe"])
def test_str_to_bool_false(string):
    assert not ford.utils.str_to_bool(string)


def test_str_to_bool_already_bool():
    assert ford.utils.str_to_bool(True)
    assert not ford.utils.str_to_bool(False)



#!/usr/bin/env python
# -- coding: utf-8 --
#
#  ford.py
#
#  Copyright 2014 Christopher MacMackin <cmacmackin@gmail.com>
#  This file is part of FORD.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from contextlib import contextmanager
from io import StringIO
import itertools
import sys
import argparse
import markdown
import os
import pathlib
import subprocess
from datetime import date, datetime
from typing import Union
from textwrap import dedent

import ford.fortran_project
import ford.sourceform
import ford.output
import ford.utils
import ford.pagetree
from ford.md_environ import EnvironExtension

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    from setuptools_scm import get_version

    __version__ = get_version(root="..", relative_to=__file__)


__appname__ = "FORD"
__author__ = "Chris MacMackin"
__credits__ = [
    "Balint Aradi",
    "Iain Barrass",
    "Izaak Beekman",
    "Jérémie Burgalat",
    "David Dickinson",
    "Gavin Huttley",
    "Harald Klimach",
    "Nick R. Papior",
    "Marco Restelli",
    "Schildkroete23",
    "Stephen J. Turnbull",
    "Jacob Williams",
    "Stefano Zhagi",
]
__license__ = "GPLv3"
__maintainer__ = "Chris MacMackin"
__status__ = "Production"


@contextmanager
def stdout_redirector(stream):
    old_stdout = sys.stdout
    sys.stdout = stream
    try:
        yield
    finally:
        sys.stdout = old_stdout


LICENSES = {
    "by": (
        '<a rel="license" href="http://creativecommons.org/licenses/by/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by/4.0/80x15.png" /></a>'
    ),
    "by-nd": (
        '<a rel="license" href="http://creativecommons.org/licenses/by-nd/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-nd/4.0/80x15.png" /></a>'
    ),
    "by-sa": (
        '<a rel="license" href="http://creativecommons.org/licenses/by-sa/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-sa/4.0/80x15.png" /></a>'
    ),
    "by-nc": (
        '<a rel="license" href="http://creativecommons.org/licenses/by-nc/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-nc/4.0/80x15.png" /></a>'
    ),
    "by-nc-nd": (
        '<a rel="license" href="http://creativecommons.org/licenses/by-nc-nd/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-nc-nd/4.0/80x15.png" /></a>'
    ),
    "by-nc-sa": (
        '<a rel="license" href="http://creativecommons.org/licenses/by-nc-sa/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-nc-sa/4.0/80x15.png" /></a>'
    ),
    "gfdl": '<a rel="license" href="http://www.gnu.org/licenses/old-licenses/fdl-1.2.en.html">GNU Free Documentation License</a>',
    "opl": '<a rel="license" href="http://opencontent.org/openpub/">Open Publication License</a>',
    "pdl": '<a rel="license" href="http://www.openoffice.org/licenses/PDL.html">Public Documentation License</a>',
    "bsd": '<a rel="license" href="http://www.freebsd.org/copyright/freebsd-doc-license.html">FreeBSD Documentation License</a>',
    "isc": '<a rel="license" href="https://opensource.org/licenses/ISC">ISC (Internet Systems Consortium) License</a>',
    "mit": '<a rel="license" href="https://opensource.org/licenses/MIT">MIT</a>',
    "": "",
}


DEFAULT_SETTINGS = {
    "alias": [],
    "author": None,
    "author_description": None,
    "author_pic": None,
    "bitbucket": None,
    "coloured_edges": False,
    "copy_subdir": [],
    "creation_date": "%Y-%m-%dT%H:%M:%S.%f%z",
    "css": None,
    "dbg": True,
    "display": ["public", "protected"],
    "doc_license": "",
    "docmark": "!",
    "docmark_alt": "*",
    "email": None,
    "encoding": "utf-8",
    "exclude": [],
    "exclude_dir": [],
    "extensions": ["f90", "f95", "f03", "f08", "f15"],
    "external": [],
    "externalize": False,
    "extra_filetypes": [],
    "extra_mods": [],
    "extra_vartypes": [],
    "facebook": None,
    "favicon": "default-icon",
    "fixed_extensions": ["f", "for", "F", "FOR"],
    "fixed_length_limit": True,
    "force": False,
    "fpp_extensions": ["F90", "F95", "F03", "F08", "F15", "F", "FOR"],
    "github": None,
    "gitlab": None,
    "gitter_sidecar": None,
    "google_plus": None,
    "graph": False,
    "graph_dir": None,
    "graph_maxdepth": "10000",
    "graph_maxnodes": "1000000000",
    "hide_undoc": False,
    "incl_src": True,
    "include": [],
    "license": "",
    "linkedin": None,
    "lower": False,
    "macro": [],
    "mathjax_config": None,
    "max_frontpage_items": 10,
    "media_dir": None,
    "output_dir": "./doc",
    "page_dir": None,
    "parallel": 1,
    "predocmark": ">",
    "predocmark_alt": "|",
    "preprocess": True,
    "preprocessor": "cpp -traditional-cpp -E -D__GFORTRAN__",
    "print_creation_date": False,
    "privacy_policy_url": None,
    "proc_internals": False,
    "project": "Fortran Program",
    "project_bitbucket": None,
    "project_download": None,
    "project_github": None,
    "project_gitlab": None,
    "project_sourceforge": None,
    "project_url": "",
    "project_website": None,
    "quiet": False,
    "revision": None,
    "search": True,
    "show_proc_parent": False,
    "sort": "src",
    "source": False,
    "src_dir": ["./src"],
    "summary": None,
    "terms_of_service_url": None,
    "twitter": None,
    "version": None,
    "warn": False,
    "website": None,
    "year": date.today().year,
}


def convert_to_bool(name, option):
    """Convert value 'option' to a bool, with a nice error message on
    failure. Expects a list from the markdown meta-data extension"""
    if len(option) > 1:
        raise ValueError(
            f"Could not convert option '{name}' to bool: expected a single value but got a list ({option})"
        )
    try:
        return ford.utils.str_to_bool(option[0])
    except ValueError:
        raise ValueError(
            f"Could not convert option '{name}' to bool: expected 'true'/'false', got: {option[0]}"
        )


def initialize():
    """
    Method to parse and check configurations of FORD, get the project's
    global documentation, and create the Markdown reader.
    """
    args = get_command_line_arguments()

    # Read in the project-file. This will contain global documentation (which
    # will appear on the homepage) as well as any information about the project
    # and settings for generating the documentation.
    proj_docs = args.project_file.read()
    directory = os.path.dirname(args.project_file.name)

    return parse_arguments(vars(args), proj_docs, directory)


def get_command_line_arguments() -> argparse.Namespace:
    """Read the command line arguments"""

    parser = argparse.ArgumentParser(
        "ford",
        description="Document a program or library written in modern Fortran. "
        "Any command-line options over-ride those specified in the project file.",
    )
    parser.add_argument(
        "project_file",
        help="file containing the description and settings for the project",
        type=argparse.FileType("r", encoding="UTF-8"),
    )
    parser.add_argument(
        "-d",
        "--src_dir",
        action="append",
        help="directories containing all source files for the project (default: ``./src``)",
    )
    parser.add_argument(
        "-p",
        "--page_dir",
        help="directory containing static pages (default: None)",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        help="directory in which to place output files (default: ``./doc``)",
    )
    parser.add_argument("-s", "--css", help="custom style-sheet for the output")
    parser.add_argument(
        "-r",
        "--revision",
        dest="revision",
        help="source code version number or revision of the project to document",
    )
    parser.add_argument(
        "--exclude",
        action="append",
        help="list of files which should not be included in the documentation",
    )
    parser.add_argument(
        "--exclude_dir",
        action="append",
        help="list of directories whose contents should not be included in the documentation",
    )
    parser.add_argument(
        "-e",
        "--extensions",
        action="append",
        help="extensions which should be scanned for documentation (default: ``f90, f95, f03, f08``)",
    )
    parser.add_argument(
        "-m",
        "--macro",
        action="append",
        help="preprocessor macros (optionally with values) to be applied to preprocessed files",
    )
    parser.add_argument(
        "-w",
        "--warn",
        dest="warn",
        action="store_true",
        default=None,
        help="display warnings for undocumented items",
    )
    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        default=None,
        help="try to continue to read files even if there are fatal errors",
    )
    parser.add_argument(
        "-g",
        "--graph",
        dest="graph",
        action="store_true",
        default=None,
        help="generate graphs for documentation output",
    )
    parser.add_argument(
        "--no-search",
        dest="search",
        action="store_false",
        default=None,
        help="don't process documentation to produce a search feature",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        default=None,
        help="suppress normal output",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s version {__version__}",
    )
    parser.add_argument(
        "--debug",
        dest="dbg",
        action="store_true",
        default=None,
        help="display traceback if fatal exception occurs and print faulty line",
    )
    parser.add_argument(
        "-I",
        "--include",
        action="append",
        help="list of directories to be searched for ``include`` files",
    )
    parser.add_argument(
        "--externalize",
        action="store_const",
        const="true",
        help="provide information about Fortran objects in a json database for "
        "other FORD projects to refer to.",
    )
    parser.add_argument(
        "-L",
        "--external_links",
        dest="external",
        action="append",
        help="""external projects to link to.
        If an entity is not found in the sources, FORD will try to look it up in
        those external projects. If those have documentation generated by FORD with
        the externalize option, a link will be placed into the documentation wherever
        this entity is referred to.
        FORD will look in the provided paths for a modules.json file.
        """,
    )

    return parser.parse_args()


def parse_arguments(
    command_line_args: dict,
    proj_docs: str,
    directory: Union[os.PathLike, str] = os.getcwd(),
):
    """Consolidates arguments from the command line and from the project
    file, and then normalises them how the rest of the code expects
    """

    try:
        import multiprocessing

        ncpus = "{0}".format(multiprocessing.cpu_count())
    except (ImportError, NotImplementedError):
        ncpus = "0"

    DEFAULT_SETTINGS["parallel"] = ncpus

    # Set up Markdown reader
    md_ext = [
        "markdown.extensions.meta",
        "markdown.extensions.codehilite",
        "markdown.extensions.extra",
        "mdx_math",
        EnvironExtension(),
    ]
    md = markdown.Markdown(
        extensions=md_ext, output_format="html", extension_configs={}
    )

    md.convert(proj_docs)
    # Remake the Markdown object with settings parsed from the project_file
    if "md_base_dir" in md.Meta:
        md_base = md.Meta["md_base_dir"][0]
    else:
        md_base = directory
    md_ext.append("markdown_include.include")
    if "md_extensions" in md.Meta:
        md_ext.extend(md.Meta["md_extensions"])
    md = markdown.Markdown(
        extensions=md_ext,
        output_format="html",
        extension_configs={"markdown_include.include": {"base_path": md_base}},
    )

    # Re-read the project file
    proj_docs = md.reset().convert(proj_docs)
    proj_data = md.Meta

    # Get the default options, and any over-rides, straightened out
    for option, default in DEFAULT_SETTINGS.items():
        args_option = command_line_args.get(option, None)
        if args_option is not None:
            proj_data[option] = args_option
        elif option in proj_data:
            # Think if there is a safe  way to evaluate any expressions found in this list
            default_type = DEFAULT_SETTINGS.get(option, None)
            if isinstance(default_type, bool):
                proj_data[option] = convert_to_bool(option, proj_data[option])
            elif not isinstance(default_type, list):
                # If it's not supposed to be a list, then it's
                # probably supposed to be a single big block of text,
                # like a description
                proj_data[option] = "\n".join(proj_data[option])
        else:
            proj_data[option] = default

    # Evaluate paths relative to project file location
    base_dir = pathlib.Path(directory).absolute()
    proj_data["base_dir"] = base_dir

    for var in [
        "page_dir",
        "output_dir",
        "graph_dir",
        "media_dir",
        "css",
        "mathjax_config",
        "src_dir",
        "exclude_dir",
        "include",
    ]:
        if proj_data[var] is None:
            continue
        if isinstance(proj_data[var], list):
            proj_data[var] = [
                ford.utils.normalise_path(base_dir, p) for p in proj_data[var]
            ]
        else:
            proj_data[var] = ford.utils.normalise_path(base_dir, proj_data[var])

    if proj_data["favicon"].strip() != DEFAULT_SETTINGS["favicon"]:
        proj_data["favicon"] = ford.utils.normalise_path(base_dir, proj_data["favicon"])

    proj_data["display"] = [item.lower() for item in proj_data["display"]]
    proj_data["creation_date"] = datetime.now().strftime(proj_data["creation_date"])
    proj_data["relative"] = proj_data["project_url"] == ""
    proj_data["extensions"] += [
        ext for ext in proj_data["fpp_extensions"] if ext not in proj_data["extensions"]
    ]
    # Parse file extensions and comment characters for extra filetypes
    extdict = {}
    for ext in proj_data["extra_filetypes"]:
        sp = ext.split()
        if len(sp) < 2:
            continue
        if len(sp) == 2:
            extdict[sp[0]] = sp[1]  # (comment_char) only
        else:
            extdict[sp[0]] = (sp[1], sp[2])  # (comment_char and lexer_str)
    proj_data["extra_filetypes"] = extdict

    # Make sure no src_dir is contained within output_dir
    for srcdir in proj_data["src_dir"]:
        # In Python 3.9+ we can use pathlib.Path.is_relative_to
        if proj_data["output_dir"] in (srcdir, *srcdir.parents):
            raise ValueError(
                f"Source directory {srcdir} is a subdirectory of output directory {proj_data['output_dir']}."
            )

    # Check that none of the docmarks are the same
    docmarks = ["docmark", "predocmark", "docmark_alt", "predocmark_alt"]
    for first, second in itertools.combinations(docmarks, 2):
        if proj_data[first] == proj_data[second] != "":
            raise ValueError(
                f"{first} ('{proj_data[first]}') and {second} ('{proj_data[second]}') are the same"
            )

    # Add gitter sidecar if specified in metadata
    if proj_data["gitter_sidecar"] is not None:
        proj_docs += """
        <script>
            ((window.gitter = {{}}).chat = {{}}).options = {{
            room: '{}'
            }};
        </script>
        <script src="https://sidecar.gitter.im/dist/sidecar.v1.js" async defer></script>
        """.format(
            proj_data["gitter_sidecar"].strip()
        )
    # Handle preprocessor:
    if proj_data["preprocess"]:
        proj_data["preprocessor"] = proj_data["preprocessor"].split()
        command = proj_data["preprocessor"] + [os.devnull]
        # Check whether preprocessor works (reading nothing from stdin)
        try:
            subprocess.run(command, check=True, capture_output=True, text=True)
        except (subprocess.CalledProcessError, OSError) as ex:
            project_file = command_line_args["project_file"].name
            exit(
                dedent(
                    f"""\
                    Error: Testing preprocessor command (`{" ".join(command)}`) failed with error:
                        {ex.stderr.strip() if isinstance(ex, subprocess.CalledProcessError) else ex}

                    If you need to preprocess files, please fix the 'preprocessor' option in '{project_file}'.
                    Otherwise, please set 'preprocess: False' in '{project_file}'"""
                )
            )
    else:
        proj_data["fpp_extensions"] = []

    # Get the correct license for project license or use value as a custom license value.
    try:
        proj_data["license"] = LICENSES[proj_data["license"].lower()]
    except KeyError:
        print(
            'Notice: license "{}" is not a recognized value, using the value as a custom license value.'.format(
                proj_data["license"]
            )
        )
    # Get the correct license for doc license(website or doc) or use value as a custom license value.
    try:
        proj_data["doc_license"] = LICENSES[proj_data["doc_license"].lower()]
    except KeyError:
        print(
            'Notice: doc_license "{}" is not a recognized value, using the value as a custom license value.'.format(
                proj_data["doc_license"]
            )
        )

    # Return project data, docs, and the Markdown reader
    md.reset()
    return (proj_data, proj_docs, md)


def main(proj_data, proj_docs, md):
    """
    Main driver of FORD.
    """
    if proj_data["relative"]:
        proj_data["project_url"] = "."
    # Parse the files in your project
    project = ford.fortran_project.Project(proj_data)
    if len(project.files) < 1:
        print(
            "Error: No source files with appropriate extension found in specified directory."
        )
        sys.exit(1)

    # Define core macros:
    ford.utils.register_macro("url = {0}".format(proj_data["project_url"]))
    ford.utils.register_macro(
        "media = {0}".format(os.path.join(proj_data["project_url"], "media"))
    )
    ford.utils.register_macro(
        "page = {0}".format(os.path.join(proj_data["project_url"], "page"))
    )

    # Register the user defined aliases:
    for alias in proj_data["alias"]:
        ford.utils.register_macro(alias)

    # Convert the documentation from Markdown to HTML. Make sure to properly
    # handle LateX and metadata.
    base_url = ".." if proj_data["relative"] else proj_data["project_url"]
    project.markdown(md, base_url)
    project.correlate()
    project.make_links(base_url)

    # Convert summaries and descriptions to HTML
    if proj_data["relative"]:
        ford.sourceform.set_base_url(".")
    if proj_data["summary"] is not None:
        proj_data["summary"] = md.convert(proj_data["summary"])
        proj_data["summary"] = ford.utils.sub_links(
            ford.utils.sub_macros(ford.utils.sub_notes(proj_data["summary"])), project
        )
    if proj_data["author_description"] is not None:
        proj_data["author_description"] = md.convert(proj_data["author_description"])
        proj_data["author_description"] = ford.utils.sub_links(
            ford.utils.sub_macros(
                ford.utils.sub_notes(proj_data["author_description"])
            ),
            project,
        )
    proj_docs_ = ford.utils.sub_links(
        ford.utils.sub_macros(ford.utils.sub_notes(proj_docs)), project
    )
    # Process any pages
    if proj_data["page_dir"] is not None:
        page_tree = ford.pagetree.get_page_tree(
            os.path.normpath(proj_data["page_dir"]), proj_data["copy_subdir"], md
        )
        print()
    else:
        page_tree = None
    proj_data["pages"] = page_tree

    # Produce the documentation using Jinja2. Output it to the desired location
    # and copy any files that are needed (CSS, JS, images, fonts, source files,
    # etc.)

    docs = ford.output.Documentation(proj_data, proj_docs_, project, page_tree)
    docs.writeout()

    if proj_data["externalize"]:
        # save FortranModules to a JSON file which then can be used
        # for external modules
        ford.utils.external(project, make=True, path=proj_data["output_dir"])

    return 0


def run():
    proj_data, proj_docs, md = initialize()

    f = StringIO() if proj_data["quiet"] else sys.stdout
    with stdout_redirector(f):
        main(proj_data, proj_docs, md)


if __name__ == "__main__":
    run()



import shutil
import sys
import os
import pathlib
import re
from urllib.parse import urlparse

import ford

from bs4 import BeautifulSoup
import pytest


HEADINGS = re.compile(r"h[1-4]")
ANY_TEXT = re.compile(r"h[1-4]|p")


def front_page_list(settings, items):
    max_frontpage_items = int(settings["max_frontpage_items"])
    return sorted(items)[:max_frontpage_items]


@pytest.fixture(scope="module")
def example_project(tmp_path_factory):
    this_dir = pathlib.Path(__file__).parent
    tmp_path = tmp_path_factory.getbasetemp() / "example"
    shutil.copytree(this_dir / "../example", tmp_path)

    with pytest.MonkeyPatch.context() as m:
        os.chdir(tmp_path)
        m.setattr(sys, "argv", ["ford", "example-project-file.md"])
        ford.run()

    with open(tmp_path / "example-project-file.md", "r") as f:
        project_file = f.read()
    settings, _, _ = ford.parse_arguments({}, project_file, tmp_path)

    doc_path = tmp_path / "doc"

    return doc_path, settings


def read_html(filename):
    with open(filename, "r") as f:
        return BeautifulSoup(f.read(), features="html.parser")


@pytest.fixture(scope="module")
def example_index(example_project):
    path, settings = example_project
    index = read_html(path / "index.html")
    return index, settings


def test_nav_bar(example_index):
    index, settings = example_index

    navbar_links = index.nav("a")
    link_names = {link.text.strip() for link in navbar_links}

    expected_pages = {
        settings["project"],
        "Source Files",
        "Block Data",
        "Modules",
        "Procedures",
        "Derived Types",
        "Programs",
        "Abstract Interfaces",
        "Contents",
    }
    assert expected_pages == link_names


def test_jumbotron(example_index):
    """This test will probably break if a different theme or HTML/CSS
    framework is used"""

    index, settings = example_index
    jumbotron = index.find("div", "jumbotron")

    jumbotron_text = (p.text for p in jumbotron("p"))
    assert settings["summary"] in jumbotron_text

    links = [link["href"] for link in jumbotron("a")]

    for location_link in [
        "project_bitbucket",
        "project_download",
        "project_github",
        "project_gitlab",
        "project_sourceforge",
        "project_website",
    ]:
        if settings[location_link] is not None:
            assert settings[location_link] in links


def test_developer_info_box(example_index):
    index, settings = example_index

    # Assume that we have something like:
    #     `<div><h2>Developer Info</h2> .. box </div>`
    developer_info = index.find(string="Developer Info").parent.parent

    dev_text = [tag.text for tag in developer_info(ANY_TEXT)]

    for expected_text in ["author", "author_description"]:
        assert settings[expected_text] in dev_text


def test_latex(example_index):
    index, settings = example_index

    tex_tags = index("script", type=re.compile("math/tex.*"))

    assert len(tex_tags) == 4


def test_source_file_links(example_index):
    index, settings = example_index

    source_files_box = index.find(ANY_TEXT, string="Source Files").parent
    source_files_list = sorted([f.text for f in source_files_box("li")])

    assert source_files_list == front_page_list(
        settings,
        [
            "ford_test_module.fpp",
            "ford_test_program.f90",
            "ford_example_type.f90",
            "ford_interfaces.f90",
            "ford_f77_example.f",
        ],
    )


def test_module_links(example_index):
    index, settings = example_index

    modules_box = index.find(ANY_TEXT, string="Modules").parent
    modules_list = sorted([f.text for f in modules_box("li")])

    assert modules_list == front_page_list(
        settings, ["test_module", "ford_example_type_mod", "interfaces"]
    )


def test_procedures_links(example_index):
    index, settings = example_index

    proceduress_box = index.find(ANY_TEXT, string="Procedures").parent
    proceduress_list = sorted([f.text for f in proceduress_box("li")])

    procedures = front_page_list(
        settings,
        [
            "decrement",
            "do_foo_stuff",
            "do_stuff",
            "increment",
            "check",
            "apply_check",
            "higher_order_unary_f",
        ],
    )
    assert proceduress_list == procedures


def test_types_links(example_index):
    index, settings = example_index

    types_box = index.find(ANY_TEXT, string="Derived Types").parent
    types_list = sorted([f.text for f in types_box("li")])

    assert types_list == sorted(["bar", "foo", "example_type"])


def test_types_type_bound_procedure(example_project):
    path, _ = example_project
    index = read_html(path / "type/example_type.html")

    bound_procedures_section = index.find("h2", string="Type-Bound Procedures").parent

    assert "This will document" in bound_procedures_section.text, "Binding summary"
    assert (
        "This binding has more documentation" in bound_procedures_section.text
    ), "Binding full docstring"
    assert (
        "Prints how many times" in bound_procedures_section.ul.text
    ), "Full procedure summary"
    assert (
        "This subroutine has more documentation" in bound_procedures_section.ul.text
    ), "Full procedure full docstring"


def test_types_constructor_summary(example_project):
    path, _ = example_project
    index = read_html(path / "type/example_type.html")

    constructor_section = index.find("h2", string="Constructor").parent

    assert "This is a constructor for our type" in constructor_section.text
    assert "This constructor has more documentation" in constructor_section.text
    assert "specific constructor" in constructor_section.ul.text
    assert "More documentation" in constructor_section.ul.text


def test_types_constructor_page(example_project):
    path, _ = example_project
    index = read_html(path / "interface/example_type.html")

    constructor_section = index.find("h2", string=re.compile("example_type")).parent

    assert "This is a constructor for our type" in constructor_section.text
    assert "This constructor has more documentation" in constructor_section.text
    assert "specific constructor" in constructor_section.text
    assert "More documentation" in constructor_section.text


def test_types_finaliser(example_project):
    path, _ = example_project
    index = read_html(path / "type/example_type.html")

    finaliser_section = index.find("h2", string="Finalization Procedures").parent

    assert "This is the finaliser" in finaliser_section.text
    assert "This finaliser has more documentation" in finaliser_section.text
    assert "Cleans up" in finaliser_section.ul.text
    assert "More documentation" in finaliser_section.ul.text


def test_graph_submodule(example_project):
    path, _ = example_project
    index = read_html(path / "module/test_submodule.html")

    graph_nodes = index.svg.find_all("g", class_="node")

    assert len(graph_nodes) == 2
    titles = sorted([node.find("text").text for node in graph_nodes])
    assert titles == sorted(["test_module", "test_submodule"])


def test_procedure_return_value(example_project):
    path, _ = example_project
    index = read_html(path / "proc/multidimension_string.html")

    retvar = index.find(string=re.compile("Return Value")).parent
    assert (
        "character(kind=kind('a'), len=4), dimension(:, :), allocatable" in retvar.text
    )


def test_info_bar(example_project):
    path, _ = example_project
    index = read_html(path / "proc/decrement.html")

    info_bar = index.find(id="info-bar")
    assert "creativecommons" in info_bar.find(id="meta-license").a["href"]
    assert "of total for procedures" in info_bar.find(id="statements").a["title"]
    assert "4 statements" in info_bar.find(id="statements").a.text

    breadcrumb = info_bar.find(class_="breadcrumb")
    assert len(breadcrumb("li")) == 3
    breadcrumb_text = [crumb.text for crumb in breadcrumb("li")]
    assert breadcrumb_text == ["ford_test_module.fpp", "test_module", "decrement"]


def test_side_panel(example_project):
    path, _ = example_project
    index = read_html(path / "program/ford_test_program.html")

    side_panel = index.find(id="sidebar")
    assert "None" not in side_panel.text

    side_panels = index.find_all(class_="panel-primary")
    # Twice as many due to the "hidden" panel that appears at the
    # bottom on mobile
    assert len(side_panels) == 2 * 4

    variables_panel = side_panels[0]
    assert len(variables_panel("a")) == 2
    assert variables_panel.a.text == "Variables"
    variables_anchor_link = variables_panel("a")[1]
    assert variables_anchor_link.text == "global_pi"
    assert (
        variables_anchor_link["href"]
        == "../program/ford_test_program.html#variable-global_pi"
    )

    subroutines_panel = side_panels[3]
    assert len(subroutines_panel("a")) == 4
    assert subroutines_panel.a.text == "Subroutines"
    subroutines_anchor_link = subroutines_panel("a")[1]
    assert subroutines_anchor_link.text == "do_foo_stuff"
    assert (
        subroutines_anchor_link["href"]
        == "../program/ford_test_program.html#proc-do_foo_stuff"
    )

    type_index = read_html(path / "type/example_type.html")
    constructor_panel = type_index.find(id="cons-1")
    assert constructor_panel.a.text == "example_type"
    assert (
        constructor_panel.a["href"]
        == "../type/example_type.html#interface-example_type"
    )
    finaliser_panel = type_index.find(id="fins-1")
    assert finaliser_panel.a.text == "example_type_finalise"
    assert (
        finaliser_panel.a["href"]
        == "../type/example_type.html#finalproc-example_type_finalise"
    )

    check_index = read_html(path / "interface/check.html")
    check_sidebar = check_index.find(id="sidebar")
    assert "None" in check_sidebar.text
    assert check_sidebar.find_all(class_="panel-primary") == []


def test_variable_lists(example_project):
    path, _ = example_project
    index = read_html(path / "program/ford_test_program.html")

    varlist = index.find(class_="varlist")
    assert "Type" in varlist.thead.text
    assert "Attributes" in varlist.thead.text
    assert "Name" in varlist.thead.text
    assert "Initial" in varlist.thead.text
    assert "Optional" not in varlist.thead.text
    assert "Intent" not in varlist.thead.text

    assert len(varlist("tr")) == 2
    assert varlist.tbody.tr.find(class_="anchor")["id"] == "variable-global_pi"
    expected_declaration = "real(kind=real64) :: global_pi = acos(-1) a global variable, initialized to the value of pi"
    declaration_no_whitespace = varlist.tbody.text.replace("\n", "").replace(" ", "")
    assert declaration_no_whitespace == expected_declaration.replace(" ", "")


def test_deprecated(example_project):
    path, _ = example_project
    index = read_html(path / "module/test_module.html")

    apply_check_box = index.find(id="proc-apply_check").parent
    assert apply_check_box.h3.span.text == "Deprecated"


def test_private_procedure_links(example_project):
    path, _ = example_project
    index = read_html(path / "type/example_type.html")

    subroutine_box = index.find("h3", string=re.compile("example_type_say"))
    assert subroutine_box.a is None


def test_public_procedure_links(example_project):
    path, _ = example_project
    index = read_html(path / "module/test_module.html")

    subroutine_box = index.find(id="proc-increment").parent
    assert subroutine_box.a is not None
    assert subroutine_box.a["href"] == "../proc/increment.html"


def test_all_internal_links_resolve(example_project):
    """Opens every HTML file, finds all relative links and checks that
    they resolve to files that actually exist. Furthermore, if the
    link has a fragment ("#something"), check that that fragment
    exists in the specified file.

    """

    path, _ = example_project
    html_files = {}

    for html in path.glob("**/*.html"):
        with open(html, "r") as f:
            html_files[html] = BeautifulSoup(f.read(), features="html.parser")

    for html, index in html_files.items():
        for a_tag in index("a"):
            link = urlparse(a_tag.get("href", ""))
            if not link.path.startswith("."):
                continue

            link_path = (html.parent / link.path).resolve()
            assert link_path.exists(), html

            if not link.fragment:
                continue

            # Check that fragments resolve too
            index2 = html_files[link_path]
            assert index2.find("a", href=re.compile(link.fragment)), html


def test_submodule_procedure_implementation_links(example_project):
    path, _ = example_project
    module_index = read_html(path / "module/test_module.html")

    interfaces_section = module_index.find("h2", string="Interfaces").parent
    check_heading = interfaces_section.ul.li
    assert "subroutine check" in check_heading.text
    implementation_link = check_heading.a
    assert implementation_link["href"].endswith("proc/check.html")

    proc_index = read_html(path / "proc/check.html")
    check_impl_heading = proc_index.find(string=re.compile("subroutine +check")).parent
    assert check_impl_heading.text.startswith("module subroutine check")
    check_interface_link = check_impl_heading.a
    assert "Interface" in check_interface_link.text
    assert check_interface_link["href"].endswith("interface/check.html")

    interface_index = read_html(path / "interface/check.html")
    check_interface_heading = interface_index.find(
        string=re.compile("subroutine +check")
    ).parent
    assert check_interface_heading.text.startswith("public module subroutine check")
    check_impl_link = check_interface_heading.a
    assert "Implementation" in check_impl_link.text
    assert check_impl_link["href"].endswith("proc/check.html")


def test_interfaces(example_project):
    path, _ = example_project
    interface_mod_page = read_html(path / "module/interfaces.html")

    # Span inside the div we actually want, but this has an id
    box_span = interface_mod_page.find(id="interface-generic_unary_f")
    assert box_span

    box = box_span.parent.parent

    list_items = box("li")
    list_item_titles = sorted([li.h3.text.strip() for li in list_items])
    assert list_item_titles == sorted(
        [
            "public pure function real_unary_f(x)",
            "public pure function higher_order_unary_f(f, n)",
            "Dummy Procedures and Procedure Pointers",
        ]
    )

    assert len(list_items[-1]("tr")) == 2



# -*- coding: utf-8 -*-
#
#  utils.py
#  This file is part of FORD.
#
#  Copyright 2015 Christopher MacMackin <cmacmackin@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


import re
import os.path
import json
import ford.sourceform
from urllib.request import urlopen, URLError
from urllib.parse import urljoin
import pathlib
from typing import Union


NOTE_TYPE = {
    "note": "info",
    "warning": "warning",
    "todo": "success",
    "bug": "danger",
    "history": "history",
}
NOTE_RE = [
    re.compile(
        r"@({})\s*(((?!@({})).)*?)@end\1\s*(</p>)?".format(
            note, "|".join(NOTE_TYPE.keys())
        ),
        re.IGNORECASE | re.DOTALL,
    )
    for note in NOTE_TYPE
] + [
    re.compile(r"@({})\s*(.*?)\s*</p>".format(note), re.IGNORECASE | re.DOTALL)
    for note in NOTE_TYPE
]
LINK_RE = re.compile(r"\[\[(\w+(?:\.\w+)?)(?:\((\w+)\))?(?::(\w+)(?:\((\w+)\))?)?\]\]")


# Dictionary for all macro definitions to be used in the documentation.
# Each key of the form |name| will be replaced by the value found in the
# dictionary in sub_macros.
_MACRO_DICT = {}


def sub_notes(docs):
    """
    Substitutes the special controls for notes, warnings, todos, and bugs with
    the corresponding div.
    """

    def substitute(match):
        ret = (
            '</p><div class="alert alert-{}" role="alert"><h4>{}</h4>'
            "<p>{}</p></div>".format(
                NOTE_TYPE[match.group(1).lower()],
                match.group(1).capitalize(),
                match.group(2),
            )
        )
        if len(match.groups()) >= 4 and not match.group(4):
            ret += "\n<p>"
        return ret

    for regex in NOTE_RE:
        docs = regex.sub(substitute, docs)
    return docs


def get_parens(line: str, retlevel: int = 0, retblevel: int = 0) -> str:
    """
    By default akes a string starting with an open parenthesis and returns the portion
    of the string going to the corresponding close parenthesis. If retlevel != 0 then
    will return when that level (for parentheses) is reached. Same for retblevel.
    """
    if len(line) == 0:
        return line
    parenstr = ""
    level = 0
    blevel = 0
    for char in line:
        if char == "(":
            level += 1
        elif char == ")":
            level -= 1
        elif char == "[":
            blevel += 1
        elif char == "]":
            blevel -= 1
        elif (
            (char.isalpha() or char == "_" or char == ":" or char == "," or char == " ")
            and level == retlevel
            and blevel == retblevel
        ):
            return parenstr
        parenstr = parenstr + char

    if level == retlevel and blevel == retblevel:
        return parenstr
    raise RuntimeError("Couldn't parse parentheses: {}".format(line))


def paren_split(sep, string):
    """
    Splits the string into pieces divided by sep, when sep is outside of parentheses.
    """
    if len(sep) != 1:
        raise ValueError("Separation string must be one character long")
    retlist = []
    level = 0
    blevel = 0
    left = 0
    for i in range(len(string)):
        if string[i] == "(":
            level += 1
        elif string[i] == ")":
            level -= 1
        elif string[i] == "[":
            blevel += 1
        elif string[i] == "]":
            blevel -= 1
        elif string[i] == sep and level == 0 and blevel == 0:
            retlist.append(string[left:i])
            left = i + 1
    retlist.append(string[left:])
    return retlist


def quote_split(sep, string):
    """
    Splits the strings into pieces divided by sep, when sep in not inside quotes.
    """
    if len(sep) != 1:
        raise ValueError("Separation string must be one character long")
    retlist = []
    squote = False
    dquote = False
    left = 0
    i = 0
    while i < len(string):
        if string[i] == '"' and not dquote:
            if not squote:
                squote = True
            elif (i + 1) < len(string) and string[i + 1] == '"':
                i += 1
            else:
                squote = False
        elif string[i] == "'" and not squote:
            if not dquote:
                dquote = True
            elif (i + 1) < len(string) and string[i + 1] == "'":
                i += 1
            else:
                dquote = False
        elif string[i] == sep and not dquote and not squote:
            retlist.append(string[left:i])
            left = i + 1
        i += 1
    retlist.append(string[left:])
    return retlist


def sub_links(string, project):
    """
    Replace links to different parts of the program, formatted as
    [[name]] or [[name(object-type)]] with the appropriate URL. Can also
    link to an item's entry in another's page with the syntax
    [[parent-name:name]]. The object type can be placed in parentheses
    for either or both of these parts.
    """
    LINK_TYPES = {
        "module": "modules",
        "extmodule": "extModules",
        "type": "types",
        "exttype": "extTypes",
        "procedure": "procedures",
        "extprocedure": "extProcedures",
        "subroutine": "procedures",
        "extsubroutine": "extProcedures",
        "function": "procedures",
        "extfunction": "extProcedures",
        "proc": "procedures",
        "extproc": "extProcedures",
        "file": "allfiles",
        "interface": "absinterfaces",
        "extinterface": "extInterfaces",
        "absinterface": "absinterfaces",
        "extabsinterface": "extInterfaces",
        "program": "programs",
        "block": "blockdata",
    }

    SUBLINK_TYPES = {
        "variable": "variables",
        "type": "types",
        "constructor": "constructor",
        "interface": "interfaces",
        "absinterface": "absinterfaces",
        "subroutine": "subroutines",
        "function": "functions",
        "final": "finalprocs",
        "bound": "boundprocs",
        "modproc": "modprocs",
        "common": "common",
    }

    def convert_link(match):
        ERR = "Warning: Could not substitute link {}. {}"
        url = ""
        name = ""
        found = False
        searchlist = []
        item = None
        # [name,obj,subname,subobj]
        if not match.group(2):
            for key, val in LINK_TYPES.items():
                searchlist.extend(getattr(project, val))
        else:
            if match.group(2).lower() in LINK_TYPES:
                searchlist.extend(getattr(project, LINK_TYPES[match.group(2).lower()]))
            else:
                print(
                    ERR.format(
                        match.group(),
                        'Unrecognized classification "{}".'.format(match.group(2)),
                    )
                )
                return match.group()

        for obj in searchlist:
            if match.group(1).lower() == obj.name.lower():
                url = obj.get_url()
                name = obj.name
                found = True
                item = obj
                break
        else:
            print(ERR.format(match.group(), '"{}" not found.'.format(match.group(1))))
            url = ""
            name = match.group(1)

        if found and match.group(3):
            searchlist = []
            if not match.group(4):
                for key, val in SUBLINK_TYPES.items():
                    if val == "constructor":
                        if getattr(item, "constructor", False):
                            searchlist.append(item.constructor)
                        else:
                            continue
                    else:
                        searchlist.extend(getattr(item, val, []))
            else:
                if match.group(4).lower() in SUBLINK_TYPES:
                    if hasattr(item, SUBLINK_TYPES[match.group(4).lower()]):
                        if match.group(4).lower() == "constructor":
                            if item.constructor:
                                searchlist.append(item.constructor)
                        else:
                            searchlist.extend(
                                getattr(item, SUBLINK_TYPES[match.group(4).lower()])
                            )
                    else:
                        print(
                            ERR.format(
                                match.group(),
                                '"{}" can not be contained in "{}"'.format(
                                    match.group(4), item.obj
                                ),
                            )
                        )
                        return match.group()
                else:
                    print(
                        ERR.format(
                            match.group(),
                            'Unrecognized classification "{}".'.format(match.group(2)),
                        )
                    )
                    return match.group()

            for obj in searchlist:
                if match.group(3).lower() == obj.name.lower():
                    url = str(url) + "#" + obj.anchor
                    name = obj.name
                    item = obj
                    break
            else:
                print(
                    ERR.format(
                        match.group(),
                        '"{0}" not found in "{1}", linking to page for "{1}" instead.'.format(
                            match.group(3), name
                        ),
                    )
                )

        if found:
            return '<a href="{}">{}</a>'.format(url, name)
        else:
            return "<a>{}</a>".format(name)

    # Get information from links (need to build an RE)
    string = LINK_RE.sub(convert_link, string)
    return string


def register_macro(string):
    """Register a new macro definition of the form ``key = value``.
    In the documentation ``|key|`` can then be used to represent value.
    If key is already defined in the list of macros an `RuntimeError`
    will be raised.

    The function returns a tuple of the form ``(value, key)``, where
    key is ``None`` if no key definition is found in the string.

    """

    if "=" not in string:
        raise RuntimeError("Error, no alias name provided for {0}".format(string))

    chunks = string.split("=", 1)
    key = "|{0}|".format(chunks[0].strip())
    val = chunks[1].strip()

    if key in _MACRO_DICT:
        # The macro is already defined. Do not overwrite it!
        # Can be ignored if the definition is the same...
        if val != _MACRO_DICT[key]:
            raise RuntimeError(
                'Could not register macro "{0}" as "{1}" because it is already defined as "{2}".'.format(
                    key, val, _MACRO_DICT[key]
                )
            )

    # Everything OK, add the macro definition to the dict.
    _MACRO_DICT[key] = val

    return (val, key)


def sub_macros(string):
    """
    Replaces macros in documentation with their appropriate values. These macros
    are used for things like providing URLs.
    """
    for key, val in _MACRO_DICT.items():
        string = string.replace(key, val)
    return string


def external(project, make=False, path="."):
    """
    Reads and writes the information needed for processing external modules.
    """
    from ford.sourceform import (
        ExternalModule,
        ExternalFunction,
        ExternalSubroutine,
        ExternalInterface,
        ExternalType,
        ExternalVariable,
        ExternalBoundProcedure,
    )

    # attributes of a module object needed for further processing
    ATTRIBUTES = [
        "pub_procs",
        "pub_absints",
        "pub_types",
        "pub_vars",
        "functions",
        "subroutines",
        "interfaces",
        "absinterfaces",
        "types",
        "variables",
        "boundprocs",
        "vartype",
        "permission",
        "generic",
    ]

    # Mapping between entity name and its type
    ENTITIES = {
        "module": ExternalModule,
        "interface": ExternalInterface,
        "type": ExternalType,
        "variable": ExternalVariable,
        "function": ExternalFunction,
        "subroutine": ExternalSubroutine,
        "boundprocedure": ExternalBoundProcedure,
    }

    def obj2dict(intObj):
        """
        Converts an object to a dictionary.
        """
        if hasattr(intObj, "external_url"):
            return None
        extDict = {
            "name": intObj.name,
            "external_url": intObj.get_url(),
            "obj": intObj.obj,
        }
        if hasattr(intObj, "proctype"):
            extDict["proctype"] = intObj.proctype
        if hasattr(intObj, "extends"):
            if isinstance(intObj.extends, ford.sourceform.FortranType):
                extDict["extends"] = obj2dict(intObj.extends)
            else:
                extDict["extends"] = intObj.extends
        for attrib in ATTRIBUTES:
            if not hasattr(intObj, attrib):
                continue

            attribute = getattr(intObj, attrib)

            if isinstance(attribute, list):
                extDict[attrib] = [obj2dict(item) for item in attribute]
            elif isinstance(attribute, dict):
                extDict[attrib] = {key: obj2dict(val) for key, val in attribute.items()}
            else:
                extDict[attrib] = str(attribute)
        return extDict

    def modules_from_local(url: pathlib.Path):
        """
        Get module information from an external project but on the
        local file system.
        Uses the io module to work in both, Python 2 and 3.
        """
        from io import open

        with open(url / "modules.json", mode="r", encoding="utf-8") as extfile:
            return json.loads(extfile.read())

    def dict2obj(extDict, url, parent=None, remote: bool = False) -> None:
        """
        Converts a dictionary to an object and immediately adds it to the project
        """
        name = extDict["name"]
        if extDict["external_url"]:
            extDict["external_url"] = extDict["external_url"].split("/", 1)[-1]
            if remote:
                external_url = urljoin(url, extDict["external_url"])
            else:
                external_url = url / extDict["external_url"]
        else:
            external_url = extDict["external_url"]

        # Look up what type of entity this is
        obj_type = extDict.get("proctype", extDict["obj"]).lower()
        # Construct the entity
        extObj: FortranBase = ENTITIES[obj_type](name, external_url, parent)
        # Now add it to the correct project list
        project_list = getattr(project, extObj._project_list)
        project_list.append(extObj)

        if obj_type == "interface":
            extObj.proctype = extDict["proctype"]
        elif obj_type == "type":
            extObj.extends = extDict["extends"]

        for key in ATTRIBUTES:
            if key not in extDict:
                continue
            if isinstance(extDict[key], list):
                tmpLs = [
                    dict2obj(item, url, extObj, remote) for item in extDict[key] if item
                ]
                setattr(extObj, key, tmpLs)
            elif isinstance(extDict[key], dict):
                tmpDict = {
                    key2: dict2obj(item, url, extObj, remote)
                    for key2, item in extDict[key].items()
                    if item
                }
                setattr(extObj, key, tmpDict)
            else:
                setattr(extObj, key, extDict[key])
        return extObj

    if make:
        # convert internal module object to a JSON database
        extModules = [obj2dict(module) for module in project.modules]
        with open(os.path.join(path, "modules.json"), "w") as modFile:
            modFile.write(json.dumps(extModules))
    else:
        # get the external modules from the external URLs
        for urldef in project.external:
            # get the external modules from the external URL
            url, short = register_macro(urldef)
            remote = re.match("https?://", url)
            try:
                if remote:
                    # Ensure the URL ends with '/' to have urljoin work as
                    # intentend.
                    if url[-1] != "/":
                        url = url + "/"
                    extModules = json.loads(
                        urlopen(urljoin(url, "modules.json")).read().decode("utf8")
                    )
                else:
                    url = pathlib.Path(url).resolve()
                    extModules = modules_from_local(url)
            except (URLError, json.JSONDecodeError) as error:
                extModules = []
                print("Could not open external URL '{}', reason: {}".format(url, error))
            # convert modules defined in the JSON database to module objects
            for extModule in extModules:
                dict2obj(extModule, url, remote=remote)


def str_to_bool(text):
    """Convert string to bool. Only takes 'true'/'false', ignoring case"""
    if isinstance(text, bool):
        return text
    if text.capitalize() == "True":
        return True
    if text.capitalize() == "False":
        return False
    raise ValueError(
        f"Could not convert string to bool: expected 'true'/'false', got '{text}'"
    )


def normalise_path(
    base_dir: pathlib.Path, path: Union[str, pathlib.Path]
) -> pathlib.Path:
    """Tidy up path, making it absolute, relative to base_dir"""
    return (base_dir / os.path.expandvars(path)).absolute()



from ford.fortran_project import Project
from ford import DEFAULT_SETTINGS
from ford.utils import normalise_path

from copy import deepcopy
from itertools import chain

import markdown
import pytest
from bs4 import BeautifulSoup


@pytest.fixture
def copy_fortran_file(tmp_path):
    def copy_file(data):
        src_dir = tmp_path / "src"
        src_dir.mkdir()
        filename = src_dir / "test.f90"
        with open(filename, "w") as f:
            f.write(data)
        settings = deepcopy(DEFAULT_SETTINGS)
        settings["src_dir"] = [src_dir]
        return settings

    return copy_file


def create_project(settings: dict):
    md_ext = [
        "markdown.extensions.meta",
        "markdown.extensions.codehilite",
        "markdown.extensions.extra",
    ]
    md = markdown.Markdown(
        extensions=md_ext, output_format="html5", extension_configs={}
    )

    project = Project(settings)
    project.markdown(md, "..")
    project.correlate()
    return project


def test_use_without_rename(copy_fortran_file):
    data = """\
    module module1
      use module2, only: routine_1
      implicit none
      private
      interface routine
          module procedure :: routine_1
      end interface routine
      public :: routine
    end module module1

    module module2
      implicit none
      private
      public :: routine_1
    contains
      subroutine routine_1(i)
        integer,intent(in) :: i
        write(*,*) i
      end subroutine routine_1
    end module module2
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)
    assert set(project.modules[0].all_procs.keys()) == {"routine_1", "routine"}


def test_use_and_rename(copy_fortran_file):
    data = """\
    module module1
      use module2, only: routine_1 => routine
      implicit none
      private
      interface routine
          module procedure :: routine_1
      end interface routine
      public :: routine
    end module module1

    module module2
      implicit none
      private
      public :: routine
    contains
      subroutine routine(i)
        integer,intent(in) :: i
        write(*,*) i
      end subroutine routine
    end module module2
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)
    assert set(project.modules[0].all_procs.keys()) == {"routine_1", "routine"}


def test_module_use_only_everything(copy_fortran_file):
    data = """\
    module default_access
      ! No access keyword
      integer :: int_public, int_private
      private :: int_private
      real :: real_public
      real, private :: real_private

      type :: type_public
        complex :: component_public
        complex, private :: component_private
      end type type_public

      type :: type_private
        character(len=1) :: string_public
        character(len=1), private :: string_private
      end type type_private

      type type_public_no_attr
      end type type_public_no_attr

      type, public :: type_public_attr
      end type type_public_attr

      type, private :: type_private_attr
      end type type_private_attr

      private :: sub_private, func_private, type_private

    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private

      integer function func_public()
      end function func_public

      integer function func_private()
      end function func_private
    end module default_access

    module use_only_everything
      use default_access, only : int_public, real_public, type_public, sub_public, func_public
    end module use_only_everything
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)

    # Double-check we're looking at the right module
    assert project.modules[1].name == "use_only_everything"
    assert set(project.modules[1].all_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[1].pub_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[1].all_types.keys()) == {
        "type_public",
    }
    assert set(project.modules[1].pub_types.keys()) == {
        "type_public",
    }
    assert set(project.modules[1].all_vars.keys()) == {
        "int_public",
        "real_public",
    }
    assert set(project.modules[1].pub_vars.keys()) == {
        "int_public",
        "real_public",
    }


def test_module_use_only_everything_change_access(copy_fortran_file):
    data = """\
    module default_access
      ! No access keyword
      integer :: int_public, int_private
      private :: int_private
      real :: real_public
      real, private :: real_private

      type :: type_public
        complex :: component_public
        complex, private :: component_private
      end type type_public

      type :: type_private
        character(len=1) :: string_public
        character(len=1), private :: string_private
      end type type_private

      type type_public_no_attr
      end type type_public_no_attr

      type, public :: type_public_attr
      end type type_public_attr

      type, private :: type_private_attr
      end type type_private_attr

      private :: sub_private, func_private, type_private

    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private

      integer function func_public()
      end function func_public

      integer function func_private()
      end function func_private
    end module default_access

    module use_only_change_access
      use default_access, only : int_public, real_public, type_public, sub_public, func_public
      private
      public :: int_public, sub_public
    end module use_only_change_access
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)

    # Double-check we're looking at the right module
    assert project.modules[1].name == "use_only_change_access"
    assert set(project.modules[1].all_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[1].pub_procs.keys()) == {
        "sub_public",
    }
    assert set(project.modules[1].all_types.keys()) == {
        "type_public",
    }
    assert set(project.modules[1].pub_types.keys()) == set()
    assert set(project.modules[1].all_vars.keys()) == {
        "int_public",
        "real_public",
    }
    assert set(project.modules[1].pub_vars.keys()) == {
        "int_public",
    }


def test_module_use_everything(copy_fortran_file):
    data = """\
    module default_access
      ! No access keyword
      integer :: int_public, int_private
      private :: int_private
      real :: real_public
      real, private :: real_private

      type :: type_public
        complex :: component_public
        complex, private :: component_private
      end type type_public

      type :: type_private
        character(len=1) :: string_public
        character(len=1), private :: string_private
      end type type_private

      type type_no_attr
      end type type_no_attr

      type, public :: type_public_attr
      end type type_public_attr

      type, private :: type_private_attr
      end type type_private_attr

      private :: sub_private, func_private, type_private

    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private

      integer function func_public()
      end function func_public

      integer function func_private()
      end function func_private
    end module default_access

    module use_everything
      use default_access
    end module use_everything
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)

    # Double-check we're looking at the right module
    assert project.modules[1].name == "use_everything"
    assert set(project.modules[1].all_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[1].pub_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[1].all_types.keys()) == {
        "type_public",
        "type_no_attr",
        "type_public_attr",
    }
    assert set(project.modules[1].pub_types.keys()) == {
        "type_public",
        "type_no_attr",
        "type_public_attr",
    }
    assert set(project.modules[1].all_vars.keys()) == {
        "int_public",
        "real_public",
    }
    assert set(project.modules[1].pub_vars.keys()) == {
        "int_public",
        "real_public",
    }


def test_module_use_everything_reexport(copy_fortran_file):
    data = """\
    module default_access
      ! No access keyword
      integer :: int_public, int_private
      private :: int_private
      real :: real_public
      real, private :: real_private

      type :: type_public
        complex :: component_public
        complex, private :: component_private
      end type type_public

      type :: type_private
        character(len=1) :: string_public
        character(len=1), private :: string_private
      end type type_private

      type type_no_attr
      end type type_no_attr

      type, public :: type_public_attr
      end type type_public_attr

      type, private :: type_private_attr
      end type type_private_attr

      private :: sub_private, func_private, type_private

    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private

      integer function func_public()
      end function func_public

      integer function func_private()
      end function func_private
    end module default_access

    module use_everything
      use default_access
    end module use_everything

    module reexport
      use use_everything
    end module reexport
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)

    # Double-check we're looking at the right module
    assert project.modules[2].name == "reexport"
    assert set(project.modules[2].all_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[2].pub_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[2].all_types.keys()) == {
        "type_public",
        "type_no_attr",
        "type_public_attr",
    }
    assert set(project.modules[2].pub_types.keys()) == {
        "type_public",
        "type_no_attr",
        "type_public_attr",
    }
    assert set(project.modules[2].all_vars.keys()) == {
        "int_public",
        "real_public",
    }
    assert set(project.modules[2].pub_vars.keys()) == {
        "int_public",
        "real_public",
    }


def test_member_in_other_module(copy_fortran_file):
    data = """\
    module module1
      use module2, only: testInterface
      type, public, abstract :: external_type
      contains
        procedure(testInterface), nopass, deferred :: testProc
      end type external_type
    end module module1

    module module2
      type, public, abstract :: internal_type
      contains
        procedure(testInterface), nopass, deferred :: testProc
      end type internal_type

      abstract interface
        pure function testInterface(input) result(output)
          real, intent(in) :: input
          real :: output
        end function testInterface
      end interface
    end module module2
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)

    module1 = project.modules[0]
    module2 = project.modules[1]

    assert (
        module1.all_types["external_type"].boundprocs[0].proto
        is module2.all_types["internal_type"].boundprocs[0].proto
    )

    assert (
        module2.all_types["internal_type"].boundprocs[0].proto.name == "testInterface"
    )


def test_display_derived_types(copy_fortran_file):
    """Check that we can set visibility attributes on types, #388"""

    data = """\
    module foo
      type no_attrs
      end type no_attrs

      type, public :: public_attr
      end type public_attr

      type, private :: private_attr
      end type private_attr
    end module foo
    """

    settings = copy_fortran_file(data)
    settings["display"] = ["public"]
    project = Project(settings)
    project.correlate()

    type_names = set([t.name for t in project.types])
    assert type_names == {"no_attrs", "public_attr"}


def test_display_derived_types_default_private(copy_fortran_file):
    """Check that we can set visibility attributes on types, #388"""

    data = """\
    module foo
      private

      type no_attrs
      end type no_attrs

      type, public :: public_attr
      end type public_attr

      type, private :: private_attr
      end type private_attr
    end module foo
    """

    settings = copy_fortran_file(data)
    settings["display"] = ["public"]
    project = create_project(settings)

    type_names = set([t.name for t in project.types])
    assert type_names == {"public_attr"}


def test_display_private_derived_types(copy_fortran_file):
    """Check that we can set visibility attributes on types, #388"""

    data = """\
    module foo
      private

      type no_attrs
      end type no_attrs

      type, public :: public_attr
      end type public_attr

      type, private :: private_attr
      end type private_attr
    end module foo
    """

    settings = copy_fortran_file(data)
    settings["display"] = ["public", "private"]
    project = create_project(settings)

    type_names = set([t.name for t in project.types])
    assert type_names == {"no_attrs", "public_attr", "private_attr"}


def test_interface_type_name(copy_fortran_file):
    """Check for shared prototype list"""
    data = """\
    module foo
      type name_t
      end type name_t

      type(name_t) :: var, var2
    end module foo
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)
    proto_names = [var.proto[0].name for var in project.modules[0].variables]
    assert proto_names == ["name_t", "name_t"]


def test_imported_interface_type_name(copy_fortran_file):
    """Check for shared prototype list"""
    data = """\
    module foo
      type name_t
      end type name_t
    end module foo

    module bar
      use foo
      type(name_t), allocatable :: var(:), var2(:)
    end module bar
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)
    proto_names = [var.proto[0].name for var in project.modules[1].variables]
    assert proto_names == ["name_t", "name_t"]


def test_abstract_interface_type_name(copy_fortran_file):
    """Check for shared prototype list"""
    data = """\
    module foo
      abstract interface
        subroutine hello
        end subroutine hello
      end interface

      procedure(hello) :: proc1, proc2
    end module foo
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)
    proto_names = [var.proto[0].name for var in project.modules[0].variables]
    assert proto_names == ["hello", "hello"]


def test_imported_abstract_interface_type_name(copy_fortran_file):
    """Check for shared prototype list"""
    data = """\
    module foo
      abstract interface
        subroutine hello
        end subroutine hello
      end interface

    contains
      integer function goodbye()
        goodbye = 1
      end function goodbye
    end module foo

    module bar
      use foo
      procedure(hello) :: proc1, proc2
      procedure(goodbye) :: proc3, proc4
    end module bar
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)
    proto_names = [var.proto[0].name for var in project.modules[1].variables]
    assert proto_names == ["hello", "hello", "goodbye", "goodbye"]


def test_display_internal_procedures(copy_fortran_file):
    """_very_ basic test of 'proc_internals' setting"""

    data = """\
    subroutine no_display_internals
    !! some docs
    integer :: local_variable
    !! This shouldn't be displayed
    end subroutine no_display_internals

    subroutine display_internals
    !! proc_internals: true
    !!
    !! some docs
    integer :: local_variable
    !! This shouldn't be displayed
    end subroutine display_internals
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)

    subroutine1 = project.procedures[0]
    subroutine2 = project.procedures[1]

    assert subroutine1.variables == []
    assert len(subroutine2.variables) == 1
    assert subroutine2.variables[0].name == "local_variable"


def test_exclude_dir(tmp_path):
    exclude_dir = tmp_path / "sub1" / "sub2"
    exclude_dir.mkdir(parents=True)
    src = tmp_path / "src"
    src.mkdir()

    with open(src / "include.f90", "w") as f:
        f.write("program foo\nend program")
    with open(exclude_dir / "exclude.f90", "w") as f:
        f.write("program bar\nend program")

    settings = deepcopy(DEFAULT_SETTINGS)
    settings["src_dir"] = [tmp_path]
    settings["exclude_dir"] = [normalise_path(tmp_path, "sub1")]
    project = Project(settings)

    program_names = {program.name for program in project.programs}
    assert program_names == {"foo"}


def test_exclude(tmp_path):
    exclude_dir = tmp_path / "sub1" / "sub2"
    exclude_dir.mkdir(parents=True)
    src = tmp_path / "src"
    src.mkdir()

    with open(src / "include.f90", "w") as f:
        f.write("program foo\nend program")
    with open(exclude_dir / "exclude.f90", "w") as f:
        f.write("program bar\nend program")

    settings = deepcopy(DEFAULT_SETTINGS)
    settings["src_dir"] = [tmp_path]
    settings["exclude"] = ["exclude.f90"]
    project = Project(settings)

    program_names = {program.name for program in project.programs}
    assert program_names == {"foo"}


@pytest.mark.parametrize(
    "sort_kind, expected_order",
    [
        (
            "src",
            {
                "variables": ["b", "a", "d", "f", "c", "e", "g", "h", "j", "i"],
                "types": ["bet", "alef", "gimel"],
                "subroutines": ["bravo", "alpha", "charlie"],
                "functions": ["beaver", "aardvark", "cat"],
                "absinterfaces": ["apple", "carrot", "banana"],
            },
        ),
        (
            "alpha",
            {
                "variables": ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"],
                "types": ["alef", "bet", "gimel"],
                "subroutines": ["alpha", "bravo", "charlie"],
                "functions": ["aardvark", "beaver", "cat"],
                "absinterfaces": ["apple", "banana", "carrot"],
            },
        ),
        (
            "permission",
            {
                "variables": ["d", "f", "c", "e", "b", "a", "g", "h", "j", "i"],
                "types": ["gimel", "alef", "bet"],
                "subroutines": ["charlie", "alpha", "bravo"],
                "functions": ["cat", "aardvark", "beaver"],
                "absinterfaces": ["apple", "carrot", "banana"],
            },
        ),
        (
            "permission-alpha",
            {
                "variables": ["d", "f", "c", "e", "a", "b", "g", "h", "i", "j"],
                "types": ["gimel", "alef", "bet"],
                "subroutines": ["charlie", "alpha", "bravo"],
                "functions": ["cat", "aardvark", "beaver"],
                "absinterfaces": ["apple", "banana", "carrot"],
            },
        ),
        (
            "type",
            {
                "variables": ["h", "g", "j", "i", "b", "a", "c", "e", "d", "f"],
                "types": ["bet", "alef", "gimel"],
                "subroutines": ["bravo", "alpha", "charlie"],
                "functions": ["cat", "beaver", "aardvark"],
                "absinterfaces": ["apple", "carrot", "banana"],
            },
        ),
        (
            "type-alpha",
            {
                "variables": ["h", "g", "i", "j", "a", "b", "c", "e", "d", "f"],
                "types": ["alef", "bet", "gimel"],
                "subroutines": ["alpha", "bravo", "charlie"],
                "functions": ["cat", "aardvark", "beaver"],
                "absinterfaces": ["apple", "banana", "carrot"],
            },
        ),
    ],
)
def test_sort(copy_fortran_file, sort_kind, expected_order):
    data = """\
    module mod_a
      type :: bet
      end type bet

      type :: alef
      end type alef

      type :: gimel
      end type gimel

      integer, private :: b, a
      type(gimel), public :: d, f
      type(alef), protected :: c, e

      character(len=2, kind=4) :: g
      character(len=1, kind=4) :: h
      character(len=1, kind=8) :: j, i
      private :: g, h, i, j

      abstract interface
        subroutine banana
        end subroutine banana

        function apple()
          import gimel
          class(gimel), allocatable :: apple
        end function apple

        type(alef) function carrot()
          import alef
        end function carrot
      end interface

      public :: gimel, charlie, cat
      protected :: alef, alpha, aardvark
      private :: bet, bravo, beaver

    contains
      subroutine bravo
      end subroutine bravo

      subroutine alpha
      end subroutine alpha

      subroutine charlie
      end subroutine charlie

      type(gimel) function beaver()
      end function beaver

      function aardvark()
        class(gimel), allocatable :: aardvark
      end function aardvark

      type(alef) function cat()
      end function cat
    end module mod_a
    """

    settings = copy_fortran_file(data)
    settings["sort"] = sort_kind
    settings["display"] = ["public", "private", "protected"]
    project = create_project(settings)
    module = project.modules[0]

    for entity_type in [
        "variables",
        "types",
        "subroutines",
        "functions",
        "absinterfaces",
    ]:
        entity_names = [entity.name for entity in getattr(module, entity_type)]
        assert entity_names == expected_order[entity_type], (sort_kind, entity_type)


def test_uses(copy_fortran_file):
    """Check that module `USE`s are matched up correctly"""

    data = """\
    module mod_a
    end module mod_a

    module mod_b
      use mod_a
    end module mod_b

    module mod_c
      use mod_a
      use mod_b
    end module mod_c

    module mod_d
      use unquoted_external_module
      use quoted_external_module
    end module mod_d
    """

    extra_mods = {
        "unquoted_external_module": "http://unquoted.example.com",
        "quoted_external_module": '"http://quoted.example.org"',
    }

    settings = copy_fortran_file(data)
    settings["extra_mods"] = [f"{key}: {value}" for key, value in extra_mods.items()]

    project = create_project(settings)
    mod_a = project.modules[0]
    mod_b = project.modules[1]
    mod_c = project.modules[2]
    mod_d = project.modules[3]

    assert mod_a.uses == set()
    assert mod_b.uses == {mod_a}
    assert mod_c.uses == {mod_a, mod_b}

    mod_d_links = list(mod_d.uses)
    for link in mod_d_links:
        soup = BeautifulSoup(link.get_url(), features="html.parser")
        link = soup.a
        expected_link = extra_mods[soup.text].strip(r"\"'")
        assert link["href"] == expected_link, link


def test_submodule_uses(copy_fortran_file):
    """Check that module `USE`s are matched up correctly"""

    data = """\
    module mod_a
    end module mod_a

    submodule (mod_a) mod_b
    end submodule mod_b

    submodule (mod_a) mod_c
    end submodule mod_c

    submodule (mod_a:mod_c) mod_d
    end submodule mod_d
    """

    extra_mods = {
        "unquoted_external_module": "http://unquoted.example.com",
        "quoted_external_module": '"http://quoted.example.org"',
    }

    settings = copy_fortran_file(data)
    settings["extra_mods"] = [f"{key}: {value}" for key, value in extra_mods.items()]

    project = create_project(settings)
    mod_a = project.modules[0]
    mod_b = project.submodules[0]
    mod_c = project.submodules[1]
    mod_d = project.submodules[2]

    assert mod_a.uses == set()

    assert mod_b.parent_submodule is None
    assert mod_b.ancestor_module == mod_a

    assert mod_c.parent_submodule is None
    assert mod_c.ancestor_module == mod_a

    assert mod_d.parent_submodule == mod_c
    assert mod_d.ancestor_module == mod_a


def test_make_links(copy_fortran_file):
    links = "[[a]] [[b]] [[b:c]] [[d]] [[b:e]] [[f]] [[a:g]] [[h]]"

    data = f"""\
    module a !! {links}
      type b !! {links}
        integer :: c !! {links}
      contains
        final :: d !! {links}
        procedure :: e !! {links}
      end type b

      interface b !! {links}
        module procedure f
      end interface b

      type(b) :: g !! {links}
    contains
      subroutine d(self) !! {links}
        type(b) :: self !! {links}
      end subroutine d
      subroutine e(self) !! {links}
        class(b) :: self !! {links}
      end subroutine e
      function f() !! {links}
        type(b) :: f !! {links}
      end function f
    end module a

    program h !! {links}
    end program h
    """
    settings = copy_fortran_file(data)
    project = create_project(settings)

    project.make_links()

    expected_links = {
        "a": "../module/a.html",
        "b": "../type/b.html",
        "c": "../type/b.html#variable-c",
        "d": "../proc/d.html",
        "e": "../type/b.html#boundprocedure-e",
        "f": "../proc/f.html",
        "g": "../module/a.html#variable-g",
        "h": "../program/h.html",
    }

    for item in chain(
        project.files[0].children,
        project.types[0].children,
        project.procedures[0].children,
        project.procedures[2].children,
    ):
        docstring = BeautifulSoup(item.doc, features="html.parser")
        link_locations = {a.string: a.get("href", None) for a in docstring("a")}

        assert link_locations == expected_links, (item, item.name)


def test_submodule_procedure_issue_446(copy_fortran_file):
    data = """\
    module foo_mod
      interface
        module subroutine quaxx
        end subroutine quaxx
      end interface
    end module foo_mod

    submodule (foo_mod) bar_submod
    contains
      module procedure quaxx
        print*, "implementation"
      end procedure
    end submodule bar_submod
    """

    settings = copy_fortran_file(data)
    settings["display"] = ["public", "private", "protected"]
    project = create_project(settings)
    module = project.modules[0]

    interface = module.interfaces[0]
    assert interface.name == "quaxx"
    submodproc = module.descendants[0].modprocedures[0]

    assert interface.procedure.module == submodproc



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# fixed2free2.py: Conversion of Fortran code from fixed to free
#                 source form.
#
# Copyright (C) 2012    Elias Rabel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Script that converts fixed form Fortran code to free form
Usage: file name as first command line parameter

python fixed2free2.py file.f > file.f90
"""

# author: Elias Rabel, 2012
# Let me know if you find this script useful:
# ylikx.0 at gmail
# https://www.github.com/ylikx/

# TODO:
# *) Improve command line usage

import sys


class FortranLine:
    def __init__(self, line, length_limit=True):
        self.line = line
        self.line_conv = line
        self.isComment = False
        self.isContinuation = False
        self.length_limit = length_limit
        self.__analyse()

    def __repr__(self):
        return self.line_conv

    def continueLine(self):
        """Insert line continuation symbol at end of line."""

        if not (self.isLong and self.is_regular):
            self.line_conv = self.line_conv.rstrip() + " &\n"
        else:
            temp = self.line_conv[:72].rstrip() + " &"
            self.line_conv = temp.ljust(72) + self.excess_line

    def __analyse(self):
        line = self.line
        firstchar = line[0] if len(line) > 0 else ""
        self.label = line[0:5].strip().lower() + " " if len(line) > 1 else ""
        cont_char = line[5] if len(line) >= 6 else ""
        fivechars = line[1:5] if len(line) > 1 else ""
        self.isShort = len(line) <= 6
        self.isLong = len(line) > 73 and self.length_limit

        self.isComment = firstchar in "cC*!"
        self.isNewComment = "!" in fivechars and not self.isComment
        self.isOMP = self.isComment and fivechars.lower() == "$omp"
        if self.isOMP:
            self.isComment = False
            self.label = ""
        self.isCppLine = firstchar == "#"
        self.is_regular = not (
            self.isComment or self.isNewComment or self.isCppLine or self.isShort
        )
        self.isContinuation = (
            not (cont_char.isspace() or cont_char == "0") and self.is_regular
        )

        if self.isLong and self.is_regular:
            self.excess_line = "!" + line[72:]
            line = line[:72] + "\n"
        else:
            self.excess_line = ""

        self.line = line
        self.__convert()

    def __convert(self):
        line = self.line
        self.code = line[6:] if len(line) > 6 else "\n"

        if self.isComment:
            self.line_conv = "!" + line[1:]
        elif self.isNewComment or self.isCppLine:
            self.line_conv = line
        elif self.isOMP:
            self.line_conv = "!" + line[1:5] + " " + self.code
        elif not self.label.isspace():
            self.line_conv = self.label + self.code
        else:
            self.line_conv = self.code

        if self.isLong and self.is_regular:
            self.line_conv = self.line_conv.rstrip().ljust(72) + self.excess_line


def convertToFree(stream, length_limit=True):
    """Convert stream from fixed source form to free source form."""
    linestack = []

    for line in stream:
        convline = FortranLine(line, length_limit)

        if convline.is_regular:
            if convline.isContinuation and linestack:
                linestack[0].continueLine()
            for line_ in linestack:
                yield str(line_)
            linestack = []

        linestack.append(convline)

    for line in linestack:
        yield str(line)


if __name__ == "__main__":
    if len(sys.argv) > 1:
        with open(sys.argv[1], "r") as infile:
            for line in convertToFree(infile):
                print(line)

    else:
        print(__doc__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  reader.py
#  This file is part of FORD.
#
#  Copyright 2014 Christopher MacMackin <cmacmackin@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import sys
import re
import ford.utils
import subprocess

if sys.version_info[0] > 2:
    from io import StringIO
else:
    from StringIO import StringIO
import os.path

from ford.fixed2free2 import convertToFree


def _contains_unterminated_string(string: str) -> bool:
    """Return True if `string` contains an unterminated quote"""
    in_quote = False
    current_quote = None
    previous_char = None
    for char in string:
        # Non-quote characters don't bother us
        if not (char == "'" or char == '"'):
            previous_char = char
            continue
        # Doubling-up a quote character doesn't make us leave a quote
        if char == previous_char:
            previous_char = char
            continue
        # If the current character is the same as the starting quote character
        # then we have left the quote (as we've dealt with doubled-quotes)
        if char == current_quote:
            in_quote = False
            current_quote = None
            previous_char = char
            continue
        # If we weren't in a quote before, we are now
        if not in_quote:
            current_quote = char
            in_quote = True
        previous_char = char
    return in_quote


def _match_docmark(docmark, line: str, in_quote: bool):
    """If docmark exists, and we're not in a string literal, try to match it"""
    if in_quote:
        return None
    if not docmark:
        return None
    return docmark.match(line)


class FortranReader(object):
    """
    An iterator which will convert a free-form Fortran source file into
    a format more conducive for analyzing. It does the following:

    - combine line continuations into one
    - remove any normal comments and any comments following an ampersand
      (line continuation)
    - if there are documentation comments preceding a piece of code, buffer
      them and return them after the code, but before any documentation
      following it
    - keep any documentation comments and, if they are at the end of a line
      of actual code, place them on a new line
    - removes blank lines and trailing white-space
    - split lines along semicolons
    """

    # Regexes
    COM_RE = re.compile("^([^\"'!]|('[^']*')|(\"[^\"]*\"))*(!.*)$")
    SC_RE = re.compile("^([^;]*);(.*)$")

    def __init__(
        self,
        filename,
        docmark="!",
        predocmark="",
        docmark_alt="",
        predocmark_alt="",
        fixed=False,
        length_limit=True,
        preprocessor=None,
        macros=[],
        inc_dirs=[],
        encoding="utf-8",
    ):
        self.name = filename

        # Check that none of the docmarks are the same
        if docmark == predocmark != "":
            raise Exception("Error: docmark and predocmark are the same.")
        if docmark == docmark_alt != "":
            raise Exception("Error: docmark and docmark_alt are the same.")
        if docmark == predocmark_alt != "":
            raise Exception("Error: docmark and predocmark_alt are the same.")
        if docmark_alt == predocmark != "":
            raise Exception("Error: docmark_alt and predocmark are the same.")
        if docmark_alt == predocmark_alt != "":
            raise Exception("Error: docmark_alt and predocmark_alt are the same.")
        if predocmark == predocmark_alt != "":
            raise Exception("Error: predocmark and predocmark_alt are the same.")

        if preprocessor:
            # Populate the macro definition and include directory path from
            # the input lists.  To define a macro we prepend '-D' and for an
            # include path we prepend '-I'.  It's important that we do not
            # prepend to an empty string as 'cpp ... -D file.F90' doesn't do
            # what is desired; use filter to remove these.
            macros = ["-D" + mac.strip() for mac in filter(None, macros)]
            incdirs = [f"-I{d}" for d in inc_dirs]
            preprocessor = preprocessor + macros + incdirs + [filename]
            command = " ".join(preprocessor)
            print(f"Preprocessing {filename}")
            try:
                out = subprocess.run(
                    preprocessor, encoding=encoding, check=True, capture_output=True
                )
                if out.stderr:
                    print(
                        f"Warning when preprocessing {filename}:\n{command}\n{out.stderr}"
                    )
                self.reader = StringIO(out.stdout)
            except subprocess.CalledProcessError as err:
                print(
                    f"Warning: error when preprocessing {filename}:\n{command}\n{err.stderr}"
                )
                print("Reverting to unpreprocessed file")
                self.reader = open(filename, "r", encoding=encoding)
        else:
            self.reader = open(filename, "r", encoding=encoding)

        if fixed:
            self.reader = convertToFree(self.reader, length_limit)

        self.fixed = fixed
        self.length_limit = length_limit
        self.inc_dirs = inc_dirs
        self.docbuffer = []
        self.pending = []
        self.prevdoc = False
        self.reading_alt = 0
        self.docmark = docmark
        self.doc_re = re.compile(
            "^([^\"'!]|('[^']*')|(\"[^\"]*\"))*(!{}.*)$".format(re.escape(docmark))
        )
        self.predocmark = predocmark
        self.encoding = encoding
        if len(self.predocmark) != 0:
            self.predoc_re = re.compile(
                "^([^\"'!]|('[^']*')|(\"[^\"]*\"))*(!{}.*)$".format(
                    re.escape(predocmark)
                )
            )
        else:
            self.predoc_re = None
        self.docmark_alt = docmark_alt
        if len(self.docmark_alt) != 0:
            self.doc_alt_re = re.compile(
                "^([^\"'!]|('[^']*')|(\"[^\"]*\"))*(!{}.*)$".format(
                    re.escape(docmark_alt)
                )
            )
        else:
            self.doc_alt_re = None
        self.predocmark_alt = predocmark_alt
        if len(self.predocmark_alt) != 0:
            self.predoc_alt_re = re.compile(
                "^([^\"'!]|('[^']*')|(\"[^\"]*\"))*(!{}.*)$".format(
                    re.escape(predocmark_alt)
                )
            )
        else:
            self.predoc_alt_re = None

        self.line_number = 0

    def __iter__(self):
        return self

    def pass_back(self, line):
        self.pending.insert(0, line)

    # for Python 2:
    def next(self):
        return self.__next__()

    def __next__(self):  # Python 3
        # If there are any lines waiting to be returned, return them
        if len(self.pending) != 0:
            self.include()
            self.prevdoc = False
            return self.pending.pop(0)
        # If there are any documentation lines waiting to be returned, return them.
        # This can happen for inline and preceding docs
        elif len(self.docbuffer) != 0:
            self.prevdoc = True
            return self.docbuffer.pop(0)

        # Loop through the source code until you have a complete line (including
        # all line continuations), or a complete preceding doc block
        done = False
        continued = False
        reading_predoc = False
        reading_predoc_alt = 0
        linebuffer = ""

        while not done:
            line = next(self.reader)

            self.line_number += 1

            in_quote = _contains_unterminated_string(linebuffer)

            if len(line.strip()) > 0 and line.strip()[0] == "#":
                continue

            # Capture any preceding documentation comments
            match = _match_docmark(self.predoc_re, line, in_quote)
            if match:
                # Switch to predoc: following comment lines are predoc until the end of the block
                reading_predoc = True
                self.reading_alt = 0
                reading_predoc_alt = 0
                # Substitute predocmark with docmark
                tmp = match.group(4)
                tmp = tmp[:1] + self.docmark + tmp[1 + len(self.predocmark) :]
                self.docbuffer.append(tmp)
                if len(line[0 : match.start(4)].strip()) > 0:
                    raise Exception(
                        "Preceding documentation lines can not be inline: {}".format(
                            line
                        )
                    )

            # Check for alternate preceding documentation
            match = _match_docmark(self.predoc_alt_re, line, in_quote)
            if match:
                # Switch to doc_alt: following comment lines are documentation until end of the block
                reading_predoc_alt = 1
                self.reading_alt = 0
                reading_predoc = False
                # Substitute predocmark_alt with docmark
                tmp = match.group(4)
                tmp = tmp[:1] + self.docmark + tmp[1 + len(self.predocmark_alt) :]
                self.docbuffer.append(tmp)
                if len(line[0 : match.start(4)].strip()) > 0:
                    raise RuntimeError(
                        f"In file {self.name}\n{self.line_number}|  {line.strip()}\n"
                        f"{'^':>{len(str(self.line_number)) + match.start(4) + 4}}\n"
                        f"Preceding alternate documentation lines can not be inline\n"
                        f"Use the docmark '!{self.docmark}' instead"
                    )

            # Check for alternate succeeding documentation
            match = _match_docmark(self.doc_alt_re, line, in_quote)
            if match:
                # Switch to doc_alt: following comment lines are documentation until end of the block
                self.reading_alt = 1
                reading_predoc = False
                reading_predoc_alt = 0
                # Substitute predocmark_alt with docmark
                tmp = match.group(4)
                tmp = tmp[:1] + self.docmark + tmp[1 + len(self.docmark_alt) :]
                self.docbuffer.append(tmp)
                if len(line[0 : match.start(4)].strip()) > 0:
                    raise Exception(
                        "Alternate documentation lines can not be inline: {}".format(
                            line
                        )
                    )

            # Capture any documentation comments
            match = _match_docmark(self.doc_re, line, in_quote)
            if match:
                self.reading_alt = 0
                reading_predoc_alt = 0
                self.docbuffer.append(match.group(4))
                line = line[0 : match.start(4)]

            if len(line.strip()) == 0 or line.strip()[0] != "!":
                self.reading_alt = 0

            if len(line.strip()) != 0 and line.strip()[0] != "!":
                reading_predoc_alt = 0

            # Remove any regular comments, unless following an alternative (pre)docmark
            match = _match_docmark(self.COM_RE, line, in_quote)
            if match:
                if (reading_predoc_alt > 1 or self.reading_alt > 1) and len(
                    line[0 : match.start(4)].strip()
                ) == 0:
                    tmp = match.group(4)
                    tmp = tmp[:1] + self.docmark + tmp[1:]
                    self.docbuffer.append(tmp)
                line = line[0 : match.start(4)]
            line = line.strip()

            # If this is a blank line following previous documentation, return
            # a line of empty documentation.
            if len(line) == 0:
                if self.prevdoc and len(self.docbuffer) == 0:
                    # ~ self.prevdoc = False
                    self.docbuffer.append("!" + self.docmark)
            else:
                reading_predoc = False
                reading_predoc_alt = 0
                self.reading_alt = 0
                # Check if line is immediate continuation of previous
                if line[0] == "&":
                    if continued:
                        line = line[1:]
                        if len(line.strip()) == 0:
                            # If the line contained only an "&" and `continued==True` then
                            # we keep going.
                            continue
                    elif len(line.strip()) == 1:
                        continue
                    else:
                        raise Exception(
                            'Can not start a new line in Fortran with "&": {}'.format(
                                line
                            )
                        )
                else:
                    linebuffer = linebuffer.strip() + " "
                # Check if line will be continued
                if line[-1] == "&":
                    continued = True
                    line = line[0:-1]
                else:
                    continued = False

            if self.reading_alt > 0:
                self.reading_alt += 1

            if reading_predoc_alt > 0:
                reading_predoc_alt += 1

            # Add this line to the buffer then check whether we're done here
            linebuffer += line
            # ~ print(((len(self.docbuffer) > 0) or (len(linebuffer) > 0)), not continued, not reading_predoc, (reading_predoc_alt == 0))
            done = (
                ((len(self.docbuffer) > 0) or (len(linebuffer) > 0))
                and not continued
                and not reading_predoc
                and (reading_predoc_alt == 0)
            )

        # Split buffer with semicolons
        # ~ print(count,linebuffer,len(linebuffer))
        frags = ford.utils.quote_split(";", linebuffer)
        self.pending.extend([s.strip() for s in frags if len(s) > 0])

        # Return the line
        if len(self.pending) > 0:
            self.include()
            self.prevdoc = False
            return self.pending.pop(0)
        else:
            tmp = self.docbuffer.pop(0)
            if tmp != "!" + self.docmark:
                self.prevdoc = True
            return tmp

    def include(self):
        """
        If the next line is an include statement, inserts the contents
        of the included file into the pending buffer.
        """
        if len(self.pending) == 0 or not self.pending[0].lower().startswith("include "):
            return
        curpending = self.pending.pop(0)
        name = curpending[8:].strip()[1:-1]
        for b in [os.path.dirname(self.name)] + self.inc_dirs:
            pname = os.path.abspath(os.path.expanduser(os.path.join(b, name)))
            if os.path.isfile(pname):
                name = pname
                break
        else:
            msg = 'Can not find include file "{}".'.format(name)
            if name.endswith(".h"):
                print("WARNING:", msg)
                # undo pop and return
                self.pending = [curpending] + self.pending
                return
            raise FileNotFoundError(msg)
        self.pending = (
            list(
                FortranReader(
                    name,
                    self.docmark,
                    self.predocmark,
                    self.docmark_alt,
                    self.predocmark_alt,
                    self.fixed,
                    self.length_limit,
                    inc_dirs=self.inc_dirs,
                    encoding=self.encoding,
                )
            )
            + self.pending
        )


if __name__ == "__main__":
    filename = sys.argv[1]
    for line in FortranReader(
        filename,
        docmark="!",
        predocmark=">",
        docmark_alt="#",
        predocmark_alt="<",
    ):
        print(line)
        continue



# -*- coding: utf-8 -*-
#
#  tipue_search.py
#  This file is part of FORD.
#
#  Copyright 2015 Christopher MacMackin <cmacmackin@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public
#  License along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


"""
Tipue Search
============

Serializes generated HTML to JSON that can be used by jQuery plugin - Tipue Search.

Adapted from the Pelican plugin by Talha Mansoor
https://github.com/getpelican/pelican-plugins/tree/master/tipue_search
"""

import os
import pathlib
import json
from bs4 import BeautifulSoup, FeatureNotFound, SoupStrainer
from codecs import open

try:
    from urlparse import urljoin
except ImportError:
    from urllib.parse import urljoin


class Tipue_Search_JSON_Generator(object):
    def __init__(self, output_path: os.PathLike, project_url: str):
        self.output_path = pathlib.Path(output_path)
        self.siteurl = project_url
        self.json_nodes = []
        self.only_text = SoupStrainer("div", id="text")
        self.only_title = SoupStrainer("title")

    def create_node(self, html, loc, meta={}):
        try:
            soup = BeautifulSoup(html, "lxml", parse_only=self.only_text)
            soup_title = BeautifulSoup(html, "lxml", parse_only=self.only_title)
        except FeatureNotFound:
            soup = BeautifulSoup(html, "html.parser", parse_only=self.only_text)
            soup_title = BeautifulSoup(html, "html.parser", parse_only=self.only_title)

        page_text = (
            soup.find("div", {"id": "text"})
            .get_text(" ", strip=True)
            .replace("\\(", "")
            .replace("\\)", "")
            .replace("\\[", "")
            .replace("\\]", "")
            .replace("$$", "")
            .replace("^", "&#94;")
        )

        # What happens if there is not a title.
        if soup_title.title is not None:
            page_title = "{0}".format(soup_title.title.string)
        else:
            page_title = ""

        # Should set default category?
        if "category" in meta:
            page_category = meta["category"]
        else:
            page_category = ""

        if self.siteurl != "":
            page_url = urljoin(self.siteurl, loc)
        else:
            page_url = loc

        node = {
            "title": page_title,
            "text": page_text,
            "tags": page_category,
            "loc": str(page_url),
        }

        self.json_nodes.append(node)

    def print_output(self):
        path = self.output_path / "tipuesearch" / "tipuesearch_content.js"

        root_node = {"pages": self.json_nodes}
        output = json.dumps(root_node, separators=(",", ":"), ensure_ascii=False)
        output = "var tipuesearch = " + output

        with open(path, "w", encoding="utf-8") as out:
            out.write(output)



"""This module just has a big list of the intrinsic functions to save space in sourceform.py
"""

INTRINSICS = [
    "abort",
    "abs",
    "abstract",
    "access",
    "achar",
    "acos",
    "acosh",
    "adjustl",
    "adjustr",
    "aimag",
    "aint",
    "alarm",
    "all",
    "allocatable",
    "allocate",
    "allocated",
    "and",
    "anint",
    "any",
    "asin",
    "asinh",
    "assign",
    "associate",
    "associated",
    "asynchronous",
    "atan",
    "atan2",
    "atanh",
    "atomic_add",
    "atomic_and",
    "atomic_cas",
    "atomic_define",
    "atomic_fetch_add",
    "atomic_fetch_and",
    "atomic_fetch_or",
    "atomic_fetch_xor",
    "atomic_or",
    "atomic_ref",
    "atomic_xor",
    "backtrace",
    "backspace",
    "bessel_j0",
    "bessel_j1",
    "bessel_jn",
    "bessel_y0",
    "bessel_y1",
    "bessel_yn",
    "bge",
    "bgt",
    "bind",
    "bit_size",
    "ble",
    "block",
    "block data",
    "blt",
    "btest",
    "c_associated",
    "c_f_pointer",
    "c_f_procpointer",
    "c_funloc",
    "c_loc",
    "c_sizeof",
    "cabs",
    "call",
    "case",
    "case default",
    "cdabs",
    "ceiling",
    "char",
    "character",
    "chdir",
    "chmod",
    "class",
    "close",
    "cmplx",
    "codimension",
    "co_broadcast",
    "co_max",
    "co_min",
    "co_reduce",
    "co_sum",
    "command_argument_count",
    "common",
    "compiler_options",
    "compiler_version",
    "complex",
    "concurrent",
    "conjg",
    "contains",
    "contiguous",
    "continue",
    "cos",
    "cosh",
    "count",
    "cpu_time",
    "critical",
    "cshift",
    "cycle",
    "data",
    "ctime",
    "dabs",
    "date_and_time",
    "dble",
    "dcmplx",
    "deallocate",
    "deferred",
    "digits",
    "dim",
    "dimension",
    "do",
    "dlog",
    "dlog10",
    "dmax1",
    "dmin1",
    "dot_product",
    "double complex",
    "double precision",
    "dprod",
    "dreal",
    "dshiftl",
    "dshiftr",
    "dsqrt",
    "dtime",
    "elemental",
    "else",
    "else if",
    "elseif",
    "elsewhere",
    "end",
    "end associate",
    "end block",
    "end block data",
    "end critical",
    "end do",
    "end enum",
    "end forall",
    "end function",
    "end if",
    "end interface",
    "end module",
    "end program",
    "end select",
    "end submodule",
    "end subroutine",
    "end type",
    "end where",
    "endfile",
    "endif",
    "entry",
    "enum",
    "enumerator",
    "eoshift",
    "epsilon",
    "equivalence",
    "erf",
    "erfc",
    "erfc_scaled",
    "etime",
    "error stop",
    "execute_command_line",
    "exit",
    "exp",
    "exponent",
    "extends",
    "extends_type_of",
    "external",
    "fget",
    "fgetc",
    "final",
    "findloc",
    "fdate",
    "floor",
    "flush",
    "fnum",
    "forall",
    "format",
    "fput",
    "fputc",
    "fraction",
    "function",
    "free",
    "fseek",
    "fstat",
    "ftell",
    "gamma",
    "generic",
    "gerror",
    "getarg",
    "get_command",
    "get_command_argument",
    "getcwd",
    "getenv",
    "get_environment_variable",
    "go to",
    "goto",
    "getgid",
    "getlog",
    "getpid",
    "getuid",
    "gmtime",
    "hostnm",
    "huge",
    "hypot",
    "iabs",
    "iachar",
    "iall",
    "iand",
    "iany",
    "iargc",
    "ibclr",
    "ibits",
    "ibset",
    "ichar",
    "idate",
    "ieee_class",
    "ieee_copy_sign",
    "ieee_get_flag",
    "ieee_get_halting_mode",
    "ieee_get_rounding_mode",
    "ieee_get_status",
    "ieee_get_underflow_mode",
    "ieee_is_finite",
    "ieee_is_nan",
    "ieee_is_negative",
    "ieee_is_normal",
    "ieee_logb",
    "ieee_next_after",
    "ieee_rem",
    "ieee_rint",
    "ieee_scalb",
    "ieee_selected_real_kind",
    "ieee_set_flag",
    "ieee_set_halting_mode",
    "ieee_set_rounding_mode",
    "ieee_set_status",
    "ieee_support_datatype",
    "ieee_support_denormal",
    "ieee_support_divide",
    "ieee_support_flag",
    "ieee_support_halting",
    "ieee_support_inf",
    "ieee_support_io",
    "ieee_support_nan",
    "ieee_support_rounding",
    "ieee_support_sqrt",
    "ieee_support_standard",
    "ieee_support_underflow_control",
    "ieee_unordered",
    "ieee_value",
    "ieor",
    "ierrno",
    "if",
    "imag",
    "image_index",
    "implicit",
    "implicit none",
    "import",
    "include",
    "index",
    "inquire",
    "int",
    "integer",
    "intent",
    "interface",
    "intrinsic",
    "int2",
    "int8",
    "ior",
    "iparity",
    "irand",
    "is",
    "is_contiguous",
    "is_iostat_end",
    "is_iostat_eor",
    "isatty",
    "ishft",
    "ishftc",
    "isnan",
    "itime",
    "kill",
    "kind",
    "lbound",
    "lcobound",
    "leadz",
    "len",
    "len_trim",
    "lge",
    "lgt",
    "link",
    "lle",
    "llt",
    "lock",
    "lnblnk",
    "loc",
    "log",
    "log_gamma",
    "log10",
    "logical",
    "long",
    "lshift",
    "lstat",
    "ltime",
    "malloc",
    "maskl",
    "maskr",
    "matmul",
    "max",
    "max0",
    "maxexponent",
    "maxloc",
    "maxval",
    "mclock",
    "mclock8",
    "merge",
    "merge_bits",
    "min",
    "min0",
    "minexponent",
    "minloc",
    "minval",
    "mod",
    "module",
    "module procedure",
    "modulo",
    "move_alloc",
    "mvbits",
    "namelist",
    "nearest",
    "new_line",
    "nint",
    "non_overridable",
    "none",
    "nopass",
    "norm2",
    "not",
    "null",
    "nullify",
    "num_images",
    "only",
    "open",
    "or",
    "operator",
    "optional",
    "pack",
    "parameter",
    "parity",
    "pass",
    "pause",
    "pointer",
    "perror",
    "popcnt",
    "poppar",
    "precision",
    "present",
    "print",
    "private",
    "procedure",
    "product",
    "program",
    "protected",
    "public",
    "pure",
    "radix",
    "ran",
    "rand",
    "random_number",
    "random_seed",
    "range",
    "rank",
    "read",
    "real",
    "recursive",
    "rename",
    "repeat",
    "reshape",
    "result",
    "return",
    "rewind",
    "rewrite",
    "rrspacing",
    "rshift",
    "same_type_as",
    "save",
    "scale",
    "scan",
    "secnds",
    "second",
    "select",
    "select case",
    "select type",
    "selected_char_kind",
    "selected_int_kind",
    "selected_real_kind",
    "sequence",
    "set_exponent",
    "shape",
    "shifta",
    "shiftl",
    "shiftr",
    "sign",
    "signal",
    "sin",
    "sinh",
    "size",
    "sizeof",
    "sleep",
    "spacing",
    "spread",
    "sqrt",
    "srand",
    "stat",
    "stop",
    "storage_size",
    "submodule",
    "subroutine",
    "sum",
    "sync all",
    "sync images",
    "sync memory",
    "symlnk",
    "system",
    "system_clock",
    "tan",
    "tanh",
    "target",
    "then",
    "this_image",
    "time",
    "time8",
    "tiny",
    "trailz",
    "transfer",
    "transpose",
    "trim",
    "ttynam",
    "type",
    "type_as",
    "ubound",
    "ucobound",
    "umask",
    "unlock",
    "unlink",
    "unpack",
    "use",
    "value",
    "verify",
    "volatile",
    "wait",
    "where",
    "while",
    "write",
    "xor",
    "zabs",
]



from markdown import Markdown, Extension
from markdown.treeprocessors import Treeprocessor


def makeExtension(**kwargs):
    return StripedTableCSSExtension(**kwargs)


class StripedTableCSSExtension(Extension):
    """Add CSS class to alternate rows in table"""

    def extendMarkdown(self, md: Markdown, *args, **kwargs):
        md.treeprocessors.register(StripedTableCSSTreeprocessor(md), "table_css", 100)
        md.registerExtension(self)


class StripedTableCSSTreeprocessor(Treeprocessor):
    def run(self, root):
        for element in root.iter():
            if element.tag == "table":
                element.set("class", "table table-striped")
